from pychess.Utils.Cord import Cord
from pychess.Utils.const import DROP, NORMAL_MOVE, PAWN, SITTUYINCHESS, QUEEN, KING, \
    WHITE, BLACK, W_OOO, W_OO, B_OOO, B_OO, QUEEN_CASTLE, FISCHERRANDOMCHESS, GATINGS,\
    KING_CASTLE, CAMBODIANCHESS, ENPASSANT, PROMOTIONS, CASTLE_SAN, C1, G1, reprSign, \
    LIGHTBRIGADECHESS, KNIGHT
from pychess.Utils.lutils.lmovegen import newMove
from .lutils import lmove


class Move:

    def __init__(self, cord0, cord1=None, board=None, promotion=None):
        """ Inits a new highlevel Move object.
            The object can be initialized in the follow ways:
                Move(cord0, cord1, board, [promotionPiece])
                Move(lovLevelMoveInt) """

        if not cord1:
            self.move = cord0
            self.flag = self.move >> 12
            self.cord0 = None if self.flag == DROP else Cord(lmove.FCORD(
                self.move))
            self.cord1 = Cord(lmove.TCORD(self.move))
        else:
            assert cord0 is not None and cord1 is not None, "cord0=%s, cord1=%s, board=%s" % (
                cord0, cord1, board)
            assert board[cord0] is not None, "cord0=%s, cord1=%s, board=%s" % (
                cord0, cord1, board)
            self.cord0 = cord0
            self.cord1 = cord1
            if not board:
                raise ValueError(
                    "Move needs a Board object in order to investigate flags")

            self.flag = NORMAL_MOVE
            if board[self.cord0].piece == PAWN and \
                    self.cord1.cord in board.PROMOTION_ZONE[board.board.color] and \
                    board.variant != SITTUYINCHESS:
                if promotion is None:
                    if board.variant == LIGHTBRIGADECHESS:
                        self.flag = lmove.FLAG_PIECE(QUEEN if board.board.color == WHITE else KNIGHT)
                    else:
                        self.flag = lmove.FLAG_PIECE(QUEEN)
                else:
                    self.flag = lmove.FLAG_PIECE(promotion)

            elif board[self.cord0].piece == PAWN and board.variant == SITTUYINCHESS:
                if cord0 == cord1:
                    # in place promotion
                    self.flag = lmove.FLAG_PIECE(QUEEN)
                elif board[self.cord1] is None and \
                        (self.cord0.cord + self.cord1.cord) % 2 == 1 and \
                        (self.cord0.cord in board.PROMOTION_ZONE[board.board.color] or board.board.pieceCount[board.color][PAWN] == 1):
                    # queen move promotion
                    self.flag = lmove.FLAG_PIECE(QUEEN)

            elif board[self.cord0].piece == KING:
                if self.cord0.x - self.cord1.x == 2 and board.variant not in (CAMBODIANCHESS, FISCHERRANDOMCHESS):
                    self.flag = QUEEN_CASTLE if self.cord0.x == 4 else KING_CASTLE
                elif self.cord0.x - self.cord1.x == -2 and board.variant not in (CAMBODIANCHESS, FISCHERRANDOMCHESS):
                    self.flag = KING_CASTLE if self.cord0.x == 4 else QUEEN_CASTLE
                elif board.variant != CAMBODIANCHESS:
                    if (abs(self.cord0.x - self.cord1.x) > 1 and self.cord1.x == C1) or (
                            board.board.ini_rooks[board.color][0] == self.cord1.cord and (
                                (board.board.color == WHITE and board.board.castling & W_OOO) or (
                            board.board.color == BLACK and board.board.castling & B_OOO))):
                        self.flag = QUEEN_CASTLE
                    elif (abs(self.cord0.x - self.cord1.x) > 1 and self.cord1.x == G1) or (
                            board.board.ini_rooks[board.color][1] == self.cord1.cord and (
                                (board.board.color == WHITE and board.board.castling & W_OO) or (
                            board.board.color == BLACK and board.board.castling & B_OO))):
                        self.flag = KING_CASTLE
            elif board[self.cord0].piece == PAWN and \
                    board[self.cord1] is None and \
                    self.cord0.x != self.cord1.x and \
                    self.cord0.y != self.cord1.y:
                self.flag = ENPASSANT

            self.move = newMove(self.cord0.cord, self.cord1.cord, self.flag)

    def _get_cords(self):
        return (self.cord0, self.cord1)

    cords = property(_get_cords)

    def _get_promotion(self):
        if self.flag in PROMOTIONS:
            return lmove.PROMOTE_PIECE(self.flag)
        return None

    promotion = property(_get_promotion)

    def __repr__(self):
        promotion = "=" + reprSign[lmove.PROMOTE_PIECE(self.flag)] \
                    if self.flag in PROMOTIONS else ""

        gate_piece = reprSign[lmove.GATE_PIECE(self.flag)].lower() if self.flag in GATINGS else ""
        if self.flag == DROP:
            piece = reprSign[lmove.FCORD(self.move)]
            return piece + "@" + str(self.cord1) + promotion
        else:
            return str(self.cord0) + str(self.cord1) + gate_piece + promotion

    def __eq__(self, other):
        if isinstance(other, Move):
            return self.move == other.move

    def __hash__(self):
        return hash(self.cords)

    def is_capture(self, board):
        return self.flag == ENPASSANT or \
            board[self.cord1] is not None and \
            self.flag != QUEEN_CASTLE and self.flag != KING_CASTLE

    def as_uci(self):
        move = "%s%s%s%s" % (self.cord0.cx, self.cord0.cy, self.cord1.cx, self.cord1.cy)
        if self.flag in PROMOTIONS:
            move += reprSign[lmove.PROMOTE_PIECE(self.flag)].lower()
        return move

# Parsers


def listToMoves(board, mstrs, type=None, validate=False, ignoreErrors=False):
    return [Move(move)
            for move in lmove.listToMoves(board.board, mstrs, type, validate,
                                          ignoreErrors)]


def parseAny(board, algnot):
    return Move(lmove.parseAny(board.board, algnot))


def parseSAN(board, san):
    """ Parse a Short/Abbreviated Algebraic Notation string """

    return Move(lmove.parseSAN(board.board, san))


def parseLAN(board, lan):
    """ Parse a Long/Expanded Algebraic Notation string """

    return Move(lmove.parseLAN(board.board, lan))


def parseFAN(board, lan):
    """ Parse a Long/Expanded Algebraic Notation string """

    return Move(lmove.parseFAN(board.board, lan))


def parseAN(board, an):
    """ Parse an Algebraic Notation string """

    return Move(lmove.parseAN(board.board, an))

# Exporters


def listToSan(board, moves):
    return lmove.listToSan(board.board, (m.move for m in moves))


def toAN(board, move, short=False, castleNotation=CASTLE_SAN):
    """ Returns a Algebraic Notation string of a move
        board should be prior to the move """

    return lmove.toAN(board.board,
                      move.move,
                      short=short,
                      castleNotation=castleNotation)


def toSAN(board, move, localRepr=False):
    """ Returns a Short/Abbreviated Algebraic Notation string of a move
        The board should be prior to the move, board2 past.
        If not board2, toSAN will not test mate """

    return lmove.toSAN(board.board, move.move, localRepr)


def toLAN(board, move):
    """ Returns a Long/Expanded Algebraic Notation string of a move
        board should be prior to the move """

    return lmove.toLAN(board.board, move.move)


def toFAN(board, move):
    """ Returns a Figurine Algebraic Notation string of a move """

    return lmove.toFAN(board.board, move.move)



import asyncio
import sys
import telnetlib
import random
import time
import platform
import subprocess
import getpass
import os
import shutil

from pychess.System.Log import log
from pychess.System.prefix import getEngineDataPrefix
from pychess.ic.icc import B_DTGR_END, B_UNIT_END

if not hasattr(asyncio.StreamReader, 'readuntil'):
    from pychess.System.readuntil import readuntil, _wait_for_data
    asyncio.StreamReader.readuntil = readuntil
    asyncio.StreamReader._wait_for_data = _wait_for_data

ENCODE = [ord(i) for i in "Timestamp (FICS) v1.0 - programmed by Henrik Gram."]
ENCODELEN = len(ENCODE)
G_RESPONSE = "\x029"
FILLER = b"1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
IAC_WONT_ECHO = b''.join([telnetlib.IAC, telnetlib.WONT, telnetlib.ECHO])

_DEFAULT_LIMIT = 2 ** 16


class CanceledException(Exception):
    pass


class ICSStreamReader(asyncio.StreamReader):
    def __init__(self, limit, loop, connected_event, name):
        asyncio.StreamReader.__init__(self, limit=limit, loop=loop)
        self.connected_event = connected_event
        self.name = name

    async def read_until(self, *untils):
        while True:
            for i, until in enumerate(untils):
                start = self._buffer.find(bytearray(until, "ascii"))
                if start >= 0:
                    self._buffer = self._buffer[start:]
                    return i
                else:
                    not_find = "read_until:%s , got:'%s'" % (until, self._buffer)
                    log.debug(not_find, extra={"task": (self.name, "raw")})

            await self._wait_for_data('read_until')


class ICSStreamReaderProtocol(asyncio.StreamReaderProtocol):
    def __init__(self, stream_reader, client_connected_cb, loop, name, timeseal):
        asyncio.StreamReaderProtocol.__init__(self, stream_reader, client_connected_cb=client_connected_cb, loop=loop)
        self.name = name
        self.timeseal = timeseal
        self.connected = False
        self.stateinfo = None

    def data_received(self, data):
        cooked = self.cook_some(data)
        self._stream_reader.feed_data(cooked)

    def cook_some(self, data):
        if not self.connected:
            log.debug(data, extra={"task": (self.name, "raw")})
            self.connected = True
            if b"FatICS" in data:
                self.FatICS = True
            elif b"puertorico.com" in data:
                self.USCN = True
                data = data.replace(IAC_WONT_ECHO, b"")
            elif b"chessclub.com" in data:
                self.ICC = True
                data = data.replace(IAC_WONT_ECHO, b"")
            elif b"Starting FICS session" in data:
                data = data.replace(IAC_WONT_ECHO, b"")
        else:
            if self.timeseal:
                data, g_count, self.stateinfo = self.decode(data, self.stateinfo)
            data = data.replace(b"\r", b"").replace(b"\x07", b"")
            # enable this only for temporary debugging
            log.debug(data, extra={"task": (self.name, "raw")})

            if self.timeseal:
                for i in range(g_count):
                    self._stream_writer.write(self.encode(bytearray(G_RESPONSE, "utf-8")) + b"\n")
        return data

    def encode(self, inbuf, timestamp=None):
        assert inbuf == b"" or inbuf[-1] != b"\n"

        if not timestamp:
            timestamp = int(time.time() * 1000 % 1e7)
        enc = inbuf + bytearray('\x18%d\x19' % timestamp, "ascii")
        padding = 12 - len(enc) % 12
        filler = random.sample(FILLER, padding)
        enc += bytearray(filler)

        buf = enc

        for i in range(0, len(buf), 12):
            buf[i + 11], buf[i] = buf[i], buf[i + 11]
            buf[i + 9], buf[i + 2] = buf[i + 2], buf[i + 9]
            buf[i + 7], buf[i + 4] = buf[i + 4], buf[i + 7]

        encode_offset = random.randrange(ENCODELEN)

        for i in range(len(buf)):
            buf[i] |= 0x80
            j = (i + encode_offset) % ENCODELEN
            buf[i] = (buf[i] ^ ENCODE[j]) - 32

        buf += bytearray([0x80 | encode_offset])
        return buf

    def decode(self, buf, stateinfo=None):
        expected_table = b"[G]\n\r"
        # TODO: add support to FatICS's new zipseal protocol when it finalizes
        # expected_table = "[G]\n\r" if not self.FatICS else "[G]\r\n"
        final_state = len(expected_table)
        g_count = 0
        result = []

        if stateinfo:
            state, lookahead = stateinfo
        else:
            state, lookahead = 0, []

        lenb = len(buf)
        idx = 0
        while idx < lenb:
            buffer_item = buf[idx]
            expected = expected_table[state]
            if buffer_item == expected:
                state += 1
                if state == final_state:
                    g_count += 1
                    lookahead = []
                    state = 0
                else:
                    lookahead.append(buffer_item)
                idx += 1
            elif state == 0:
                result.append(buffer_item)
                idx += 1
            else:
                result.extend(lookahead)
                lookahead = []
                state = 0

        return bytearray(result), g_count, (state, lookahead)


# You can get ICC timestamp from
# https://www.chessclub.com/user/resources/icc/timestamp/
if sys.platform == "win32":
    timestamp = "timestamp_win32.exe"
else:
    timestamp = "timestamp_linux_2.6.8"

altpath = getEngineDataPrefix()
timestamp_path = shutil.which(timestamp, os.X_OK, path=altpath)


class ICSTelnet():
    sensitive = False

    def __init__(self, timeseal):
        self.name = ""
        self.connected = False
        self.canceled = False
        self.FatICS = False
        self.USCN = False
        self.ICC = False
        self.timeseal = timeseal
        self.ICC_buffer = ""

    async def start(self, host, port, connected_event):
        if self.canceled:
            raise CanceledException()
        self.port = port
        self.host = host
        self.connected_event = connected_event

        self.name = host

        if host == "chessclub.com":
            self.ICC = True

            if self.timeseal and timestamp_path is not None:
                self.host = "127.0.0.1"
                self.port = 5500
                try:
                    if sys.platform == "win32":
                        # To prevent engines opening console window
                        startupinfo = subprocess.STARTUPINFO()
                        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                    else:
                        startupinfo = None
                    create = asyncio.create_subprocess_exec(* ["%s" % timestamp_path, "-p", "%s" % self.port], startupinfo=startupinfo)
                    self.timestamp_proc = await create
                    log.info("%s started OK" % timestamp_path)
                except OSError as err:
                    log.info("Can't start %s OSError: %s %s" % (timestamp_path, err.errno, err.strerror))
                    self.port = port
                    self.host = host
            else:
                log.info("%s not found" % timestamp_path)

            self.timeseal = False

        def cb(reader, writer):
            reader.stream_writer = writer
            reader.connected_event.set()

        loop = asyncio.get_event_loop()
        self.reader = ICSStreamReader(_DEFAULT_LIMIT, loop, self.connected_event, self.name)
        self.protocol = ICSStreamReaderProtocol(self.reader, cb, loop, self.name, self.timeseal)
        coro = loop.create_connection(lambda: self.protocol, self.host, self.port)
        self.transport, _protocol = await coro
        # writer = asyncio.StreamWriter(transport, protocol, reader, loop)

        if self.timeseal:
            self.write(self.get_init_string())

    def cancel(self):
        self.canceled = True
        self.close()

    def close(self):
        if self.protocol.connected:
            self.protocol.connected = False
            self.transport.close()

    def get_init_string(self):
        """ timeseal header: TIMESTAMP|bruce|Linux gruber 2.6.15-gentoo-r1 #9
            PREEMPT Thu Feb 9 20:09:47 GMT 2006 i686 Intel(R) Celeron(R) CPU
            2.00GHz GenuineIntel GNU/Linux| 93049 """
        user = getpass.getuser()
        uname = ' '.join(list(platform.uname()))
        return "TIMESTAMP|" + user + "|" + uname + "|"

    def write(self, string):
        logstr = "*" * len(string) if self.sensitive else string
        self.sensitive = False
        log.info(logstr, extra={"task": (self.name, "raw")})

        if self.timeseal:
            self.transport.write(self.protocol.encode(bytearray(string, "utf-8")) + b"\n")
        else:
            self.transport.write(string.encode() + b"\n")

    async def readline(self):
        if self.canceled:
            raise CanceledException()

        if self.ICC:
            line = await self.readuntil(b"\n")
            return line.strip()
        else:
            line = await self.reader.readline()
            return line.decode("latin_1").strip()

    async def readuntil(self, until):
        if self.canceled:
            raise CanceledException()

        if self.ICC:
            if len(self.ICC_buffer) == 0:
                self.ICC_buffer = await self.reader.readuntil(until)
            i = self.ICC_buffer.find(until)
            m = sys.maxsize
            if i >= 0:
                m = i
            j = self.ICC_buffer.find(B_DTGR_END)
            if j >= 0:
                m = min(m, j)
            k = self.ICC_buffer.find(B_UNIT_END)
            if k >= 0:
                m = min(m, k)
            if m != sys.maxsize:
                if m == i:
                    stuff = self.ICC_buffer[:m + len(until)]
                    self.ICC_buffer = self.ICC_buffer[m + len(until):]
                    return stuff.decode("latin_1")
                else:
                    stuff = self.ICC_buffer[:m + 2]
                    self.ICC_buffer = self.ICC_buffer[m + 2:]
                    return stuff.decode("latin_1")
            else:
                return ""
        else:
            data = await self.reader.readuntil(until)
            return data.decode("latin_1")

    async def read_until(self, *untils):
        if self.canceled:
            raise CanceledException()

        ret = await self.reader.read_until(*untils)
        return ret



# Chess960 (Fischer Random Chess)

from random import randrange
from math import floor

from pychess.Utils.const import FISCHERRANDOMCHESS, VARIANTS_SHUFFLE
from pychess.Utils.Board import Board
from pychess.Utils.const import reprFile


class FischerandomBoard(Board):
    variant = FISCHERRANDOMCHESS
    __desc__ = _(
        "http://en.wikipedia.org/wiki/Chess960\n" +
        "FICS wild/fr: http://www.freechess.org/Help/HelpFiles/wild.html")
    name = _("Fischer Random")
    cecp_name = "fischerandom"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.shuffle_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def getFrcFen(self, position=519):
        position = max(1, min(960, position))
        pieces = ['', '', '', '', '', '', '', '']

        # Bishops
        pieces[floor(0.08 * (floor(25 * (position - 1)) % 100) + 1.5)] = 'b'
        pieces[floor(0.08 * (floor(25 * floor((position - 1) / 4)) % 100) + 0.5)] = 'b'

        # Queen
        z = floor(floor((position - 1) / 4) / 4) / 6
        p = floor(6 * (z - floor(z)) + 0.5)
        for i in range(8):
            if pieces[i] == '':
                if p == 0:
                    pieces[i] = 'q'
                    break
                p -= 1

        # KRN
        krn = ['nnrkr', 'nrnkr', 'nrknr', 'nrkrn', 'rnnkr', 'rnknr', 'rnkrn', 'rknnr', 'rknrn', 'rkrnn'][floor(z)]
        for i in range(8):
            if pieces[i] == '':
                pieces[i] = krn[:1]
                krn = krn[1:]

        # Castling
        castling = ''
        for i in range(8):
            if pieces[i] == 'r':
                castling += reprFile[i]

        # FEN
        pieces = ''.join(pieces)
        return '%s/pppppppp/8/8/8/8/PPPPPPPP/%s w %s%s - 0 1' % (pieces, pieces.upper(), castling.upper(), castling)

    def shuffle_start(self):
        return self.getFrcFen(randrange(1, 960))


if __name__ == '__main__':
    frcBoard = FischerandomBoard(True)
    for i in range(10):
        print(frcBoard.shuffle_start())



""" Wildcastle shuffle Chess """

import random

from pychess.Utils.const import WILDCASTLESHUFFLECHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board


class WildcastleShuffleBoard(Board):
    variant = WILDCASTLESHUFFLECHESS
    __desc__ = _("xboard wildcastle http://home.hccnet.nl/h.g.muller/engine-intf.html#8\n" +
                 "FICS wild/1: http://www.freechess.org/Help/HelpFiles/wild.html\n" +
                 "* In this variant both sides have the same set of pieces as in normal chess.\n" +
                 "* The white king starts on d1 or e1 and the black king starts on d8 or e8,\n" +
                 "* and the rooks are in their usual positions.\n" +
                 "* Bishops are always on opposite colors.\n" +
                 "* Subject to these constraints the position of the \
                 pieces on their first ranks is random.\n" +
                 "* Castling is done similarly to normal chess:\n" +
                 "* o-o-o indicates long castling and o-o short castling.")
    name = _("Wildcastle shuffle")
    cecp_name = "wildcastle"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.shuffle_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def shuffle_start(self):
        def get_shuffle():
            positions = [2, 3, 4, 5, 6, 7]
            back_rank = ['r'] + ([''] * 6) + ['r']

            king = random.choice((4, 5))
            back_rank[king - 1] = 'k'
            positions.remove(king)

            bishop = random.choice(positions)
            back_rank[bishop - 1] = 'b'
            positions.remove(bishop)
            color = bishop % 2

            bishop = random.choice([i for i in positions if i % 2 != color])
            back_rank[bishop - 1] = 'b'
            positions.remove(bishop)

            queen = random.choice(positions)
            back_rank[queen - 1] = 'q'
            positions.remove(queen)

            knight = random.choice(positions)
            back_rank[knight - 1] = 'n'
            positions.remove(knight)

            knight = random.choice(positions)
            back_rank[knight - 1] = 'n'
            positions.remove(knight)

            return ''.join(back_rank)

        fen = get_shuffle() + '/pppppppp/8/8/8/8/PPPPPPPP/' + \
            get_shuffle().upper() + ' w KQkq - 0 1'

        return fen



""" Rook Odds Variant"""
from pychess.Utils.const import ROOKODDSCHESS, VARIANTS_ODDS
from pychess.Utils.Board import Board

ROOKODDSSTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/1NBQKBNR w Kkq - 0 1"


class RookOddsBoard(Board):
    """:Description: Standard chess rules apply, but one side starts with one less rook"""
    variant = ROOKODDSCHESS
    __desc__ = _("One player starts with one less rook piece")
    name = _("Rook odds")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_ODDS

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=ROOKODDSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import subprocess
import os.path


class Command():
    def __init__(self, command, inputstr):
        self.command = command
        self.inputstr = inputstr

    def run(self, timeout=None):
        """ Run a command then return: (status, output, error). """
        status = None
        output = ""
        error = ""
        try:
            process = subprocess.Popen(self.command,
                                       universal_newlines=True,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       bufsize=0,
                                       cwd=os.path.dirname(self.command[0]))
        except OSError:
            return status, output, error
        except ValueError:
            return status, output, error

        try:
            output, error = process.communicate(
                input=self.inputstr,
                timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            output, error = process.communicate()

        status = process.returncode
        return status, output, error



# from: http://cairographics.org/freetypepython/

import ctypes as ct
import cairo


class PycairoContext(ct.Structure):
    _fields_ = \
        [
            ("PyObject_HEAD", ct.c_byte * object.__basicsize__),
            ("ctx", ct.c_void_p),
            ("base", ct.c_void_p),
        ]


_initialized = False


def create_cairo_font_face_for_file(filename, faceindex=0, loadoptions=0):
    """ Given the name of a font file, and optional faceindex to pass to FT_New_Face
        and loadoptions to pass to cairo_ft_font_face_create_for_ft_face, creates
        a cairo.FontFace object that may be used to render text with that font. """
    global _initialized
    global _freetype_so
    global _cairo_so
    global _ft_lib
    global _ft_destroy_key
    global _surface

    CAIRO_STATUS_SUCCESS = 0
    FT_Err_Ok = 0

    if not _initialized:
        # find shared objects
        _freetype_so = ct.CDLL("libfreetype.so.6")
        _cairo_so = ct.CDLL("libcairo.so.2")
        _cairo_so.cairo_ft_font_face_create_for_ft_face.restype = ct.c_void_p
        _cairo_so.cairo_ft_font_face_create_for_ft_face.argtypes = [ct.c_void_p, ct.c_int]
        _cairo_so.cairo_font_face_get_user_data.restype = ct.c_void_p
        _cairo_so.cairo_font_face_get_user_data.argtypes = (ct.c_void_p, ct.c_void_p)
        _cairo_so.cairo_font_face_set_user_data.argtypes = (ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_void_p)
        _cairo_so.cairo_set_font_face.argtypes = [ct.c_void_p, ct.c_void_p]
        _cairo_so.cairo_font_face_status.argtypes = [ct.c_void_p]
        _cairo_so.cairo_font_face_destroy.argtypes = (ct.c_void_p,)
        _cairo_so.cairo_status.argtypes = [ct.c_void_p]
        # initialize freetype
        _ft_lib = ct.c_void_p()
        status = _freetype_so.FT_Init_FreeType(ct.byref(_ft_lib))
        if status != FT_Err_Ok:
            raise RuntimeError("Error %d initializing FreeType library." % status)

        _surface = cairo.ImageSurface(cairo.FORMAT_A8, 0, 0)
        _ft_destroy_key = ct.c_int()  # dummy address
        _initialized = True

    ft_face = ct.c_void_p()
    cr_face = None
    try:
        # load FreeType face
        status = _freetype_so.FT_New_Face(_ft_lib, filename.encode("utf-8"), faceindex, ct.byref(ft_face))
        if status != FT_Err_Ok:
            raise RuntimeError("Error %d creating FreeType font face for %s" % (status, filename))

        # create Cairo font face for freetype face
        cr_face = _cairo_so.cairo_ft_font_face_create_for_ft_face(ft_face, loadoptions)
        status = _cairo_so.cairo_font_face_status(cr_face)
        if status != CAIRO_STATUS_SUCCESS:
            raise RuntimeError("Error %d creating cairo font face for %s" % (status, filename))

        # Problem: Cairo doesn't know to call FT_Done_Face when its font_face object is
        # destroyed, so we have to do that for it, by attaching a cleanup callback to
        # the font_face. This only needs to be done once for each font face, while
        # cairo_ft_font_face_create_for_ft_face will return the same font_face if called
        # twice with the same FT Face.
        # The following check for whether the cleanup has been attached or not is
        # actually unnecessary in our situation, because each call to FT_New_Face
        # will return a new FT Face, but we include it here to show how to handle the
        # general case.
        if _cairo_so.cairo_font_face_get_user_data(cr_face, ct.byref(_ft_destroy_key)) is None:
            status = _cairo_so.cairo_font_face_set_user_data(
                cr_face,
                ct.byref(_ft_destroy_key),
                ft_face,
                _freetype_so.FT_Done_Face)
            if status != CAIRO_STATUS_SUCCESS:
                raise RuntimeError("Error %d doing user_data dance for %s" % (status, filename))
            ft_face = None  # Cairo has stolen my reference

        # set Cairo font face into Cairo context
        cairo_ctx = cairo.Context(_surface)
        cairo_t = PycairoContext.from_address(id(cairo_ctx)).ctx
        _cairo_so.cairo_set_font_face(cairo_t, cr_face)
        status = _cairo_so.cairo_font_face_status(cairo_t)
        if status != CAIRO_STATUS_SUCCESS:
            raise RuntimeError("Error %d creating cairo font face for %s" % (status, filename))

    finally:
        _cairo_so.cairo_font_face_destroy(cr_face)
        _freetype_so.FT_Done_Face(ft_face)

    # get back Cairo font face as a Python object
    face = cairo_ctx.get_font_face()
    return face


if __name__ == '__main__':
    face = create_cairo_font_face_for_file("../../../pieces/ttf/harlequin.ttf",
                                           0)

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 200, 128)

    ctx = cairo.Context(surface)

    ctx.set_font_face(face)
    ctx.set_font_size(30)
    ctx.move_to(0, 44)
    ctx.show_text("pnbrqk")

    ctx.move_to(0, 74)
    ctx.show_text("omvtwl")

    del ctx

    surface.write_to_png("0pieces.png")



#!/usr/bin/env python

import json
import os
import re

import pexpect
from pexpect.popen_spawn import PopenSpawn

from pychess.Players.ProtocolEngine import TIME_OUT_SECOND

PGN_HEADERS_REGEX = re.compile(r"\[([A-Za-z0-9_]+)\s+\"(.*)\"\]")


class Parser:
    def __init__(self, engine=''):
        if not engine:
            engine = './parser'
        self.p = PopenSpawn(engine, timeout=TIME_OUT_SECOND, encoding="utf-8")
        self.pgn = ''
        self.db = ''

    def wait_ready(self):
        self.p.sendline('isready')
        self.p.expect(u'readyok')

    def open(self, pgn, full=True):
        '''Open a PGN file and create an index if not exsisting'''
        if not os.path.isfile(pgn):
            raise NameError("File {} does not exsist".format(pgn))
        pgn = os.path.normcase(pgn)
        self.pgn = pgn
        self.db = os.path.splitext(pgn)[0] + '.bin'
        if not os.path.isfile(self.db):
            result = self.make(full)
            self.db = result['Book file']

    def close(self):
        '''Terminate chess_db. Not really needed: engine will terminate as
           soon as pipe is closed, i.e. when we exit.'''
        self.p.sendline('quit')
        self.p.expect(pexpect.EOF)
        self.pgn = ''
        self.db = ''

    def make(self, full=True):
        '''Make an index out of a pgn file'''
        if not self.pgn:
            raise NameError("Unknown DB, first open a PGN file")
        cmd = 'book ' + self.pgn
        if full:
            cmd += ' full'
        self.p.sendline(cmd)
        self.wait_ready()
        s = '{' + self.p.before.split('{')[1]
        s = s.replace('\\', r'\\')  # Escape Windows's path delimiter
        result = json.loads(s)
        self.p.before = ''
        return result

    def find(self, fen, limit=10, skip=0):
        '''Find all games with positions equal to fen'''
        if not self.db:
            raise NameError("Unknown DB, first open a PGN file")
        cmd = "find {} limit {} skip {} {}".format(self.db, limit, skip, fen)
        self.p.sendline(cmd)
        self.wait_ready()
        result = json.loads(self.p.before)
        self.p.before = ''
        return result

    def get_games(self, list):
        '''Retrieve the PGN games specified in the offset list'''
        if not self.pgn:
            raise NameError("Unknown DB, first open a PGN file")
        pgn = []
        with open(self.pgn, "r") as f:
            for ofs in list:
                f.seek(ofs)
                game = ''
                for line in f:
                    if line.startswith('[Event "'):
                        if game:
                            break  # Second one, start of next game
                        else:
                            game = line  # First occurence
                    elif game:
                        game += line
                pgn.append(game.strip())
        return pgn

    def get_header(self, pgn):
        '''Return a dict with just header information out of a pgn game. The
           pgn tags are supposed to be consecutive'''
        header = {}
        for line in pgn.splitlines():
            line = line.strip()
            if line.startswith('[') and line.endswith(']'):
                tag_match = PGN_HEADERS_REGEX.match(line)
                if tag_match:
                    header[tag_match.group(1)] = tag_match.group(2)
                else:
                    break
        return header

    def get_game_headers(self, list):
        '''Return a list of headers out of a list of pgn games'''
        headers = []
        for pgn in list:
            h = self.get_header(pgn)
            headers.append(h)
        return headers



# -*- coding: UTF-8 -*-

from math import pi, cos, sin

import cairo
from gi.repository import GLib, Gtk, Gdk, Pango, PangoCairo, GObject

from pychess.System import conf
from pychess.Utils import formatTime
from pychess.Utils.const import BLACK, WHITE, LOCAL, UNFINISHED_STATES, DRAW, WHITEWON, BLACKWON, UNKNOWN_STATE
from . import preferencesDialog


class ChessClock(Gtk.DrawingArea):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.connect("draw", self.expose)
        self.names = [_("White"), _("Black")]

        self.model = None
        self.short_on_time = [False, False]
        self.alarm_spin = conf.get("alarm_spin")

        conf.notify_add("alarm_spin", self.on_alarm_spin)

    def on_alarm_spin(self, *args):
        self.alarm_spin = conf.get("alarm_spin")

    def expose(self, widget, ctx):
        context = widget.get_window().cairo_create()
        clip_ext = context.clip_extents()

        if clip_ext[0] > 0 or clip_ext[2] < self.get_allocated_width():
            self.redraw_canvas()
            return False

        rec = Gdk.Rectangle()
        rec.x, rec.y, rec.width, rec.height = clip_ext[0], clip_ext[1], \
            clip_ext[2] - clip_ext[0], clip_ext[3] - clip_ext[1]
        context.rectangle(rec.x, rec.y, rec.width, rec.height)
        context.clip()
        self.draw(context)
        return False

    def draw(self, context):

        style_ctxt = self.get_style_context()
        self.light = style_ctxt.lookup_color("p_light_color")[1]
        self.dark = style_ctxt.lookup_color("p_dark_color")[1]
        if not self.model:
            return

        # Draw graphical Clock. Should this be moved to preferences?
        drawClock = True

        rect = Gdk.Rectangle()
        clip_ext = context.clip_extents()
        rect.x, rect.y, rect.width, rect.height = clip_ext[0], clip_ext[1], \
            clip_ext[2] - clip_ext[0], clip_ext[3] - clip_ext[1]
        context.rectangle(rect.width / 2. * self.model.movingColor, 0,
                          rect.width / 2., rect.height)

        context.set_source_rgba(self.dark.red, self.dark.green, self.dark.blue,
                                self.dark.alpha)
        context.fill_preserve()
        context.new_path()

        time0 = self.names[0], self.formatedCache[WHITE]
        layout0 = self.create_pango_layout(" %s: %s " % (time0))
        layout0.set_font_description(Pango.FontDescription("Sans Serif 17"))

        time1 = self.names[1], self.formatedCache[BLACK]
        layout1 = self.create_pango_layout(" %s: %s " % (time1))
        layout1.set_font_description(Pango.FontDescription("Sans Serif 17"))

        dbl_max = max(layout1.get_pixel_size()[0], layout0.get_pixel_size()[0]) * 2
        self.set_size_request(dbl_max + rect.height + 7, -1)

        pangoScale = float(Pango.SCALE)

        # Analog clock code.
        def paintClock(player):
            clock_y = rect.height / 2.
            clock_x = clock_y + rect.width / 2. * player + 1
            rec = rect.height / 2. - 3.5

            context.arc(clock_x, clock_y, rec - 1, 0, 2 * pi)
            linear = cairo.LinearGradient(clock_x - rec * 2, clock_y - rec * 2,
                                          clock_x + rec * 2, clock_y + rec * 2)
            linear.add_color_stop_rgba(0, 1, 1, 1, 0.3)
            linear.add_color_stop_rgba(1, 0, 0, 0, 0.3)
            # context.set_source_rgba( 0, 0, 0, .3)
            context.set_source(linear)
            context.fill()

            linear = cairo.LinearGradient(clock_x - rec, clock_y - rec,
                                          clock_x + rec, clock_y + rec)
            linear.add_color_stop_rgba(0, 0, 0, 0, 0.5)
            linear.add_color_stop_rgba(1, 1, 1, 1, 0.5)
            context.arc(clock_x, clock_y, rec, 0, 2 * pi)
            context.set_source(linear)
            context.set_line_width(2.5)
            context.stroke()

            starttime = float(self.model.getInitialTime()) or 1
            used = self.model.getPlayerTime(player) / starttime
            if used > 0:
                if used > 0:
                    context.arc(clock_x, clock_y, rec - .8, -(used + 0.25) * 2 * pi, -0.5 *
                                pi)
                    context.line_to(clock_x, clock_y)
                    context.close_path()
                elif used == 0:
                    context.arc(clock_x, clock_y, rec - .8, -0.5 * pi, 1.5 * pi)
                    context.line_to(clock_x, clock_y)

                radial = cairo.RadialGradient(clock_x, clock_y, 3, clock_x, clock_y, rec)
                if player == 0:
                    # radial.add_color_stop_rgb(0, .73, .74, .71)
                    radial.add_color_stop_rgb(0, .93, .93, .92)
                    radial.add_color_stop_rgb(1, 1, 1, 1)
                else:
                    # radial.add_color_stop_rgb(0, .53, .54, .52)
                    radial.add_color_stop_rgb(0, .18, .20, .21)
                    radial.add_color_stop_rgb(1, 0, 0, 0)
                context.set_source(radial)
                context.fill()

                x_loc = clock_x - cos((used - 0.25) * 2 * pi) * (rec - 1)
                y_loc = clock_y + sin((used - 0.25) * 2 * pi) * (rec - 1)
                context.move_to(clock_x, clock_y - rec + 1)
                context.line_to(clock_x, clock_y)
                context.line_to(x_loc, y_loc)
                context.set_line_width(0.2)
                if player == 0:
                    context.set_source_rgb(0, 0, 0)
                else:
                    context.set_source_rgb(1, 1, 1)
                context.stroke()

        if drawClock:
            paintClock(WHITE)
        if (self.model.movingColor or WHITE) == WHITE:
            context.set_source_rgba(self.light.red, self.light.green,
                                    self.light.blue, self.light.alpha)
        else:
            context.set_source_rgba(self.dark.red, self.dark.green,
                                    self.dark.blue, self.dark.alpha)
        y_loc = rect.height / 2. - layout0.get_extents()[0].height / pangoScale / 2 \
            - layout0.get_extents()[0].y / pangoScale
        context.move_to(rect.height - 7, y_loc)
        PangoCairo.show_layout(context, layout0)

        if drawClock:
            paintClock(BLACK)
        if self.model.movingColor == BLACK:
            context.set_source_rgba(self.light.red, self.light.green,
                                    self.light.blue, self.light.alpha)
        else:
            context.set_source_rgba(self.dark.red, self.dark.green,
                                    self.dark.blue, self.dark.alpha)
        y_loc = rect.height / 2. - layout0.get_extents()[0].height / pangoScale / 2 \
            - layout0.get_extents()[0].y / pangoScale
        context.move_to(rect.width / 2. + rect.height - 7, y_loc)
        PangoCairo.show_layout(context, layout1)

    def redraw_canvas(self):
        def do_redraw_canvas():
            if self.get_window():
                allocation = self.get_allocation()
                rect = Gdk.Rectangle()
                rect.x, rect.y, rect.width, rect.height = (0, 0, allocation.width,
                                                           allocation.height)
                self.get_window().invalidate_rect(rect, True)
                self.get_window().process_updates(True)

        GLib.idle_add(do_redraw_canvas)

    def setModel(self, model):
        self.model = model
        self.model.connect("time_changed", self.time_changed)
        self.model.connect("player_changed", self.player_changed)
        self.formatedCache = [formatTime(self.model.getPlayerTime(
            self.model.movingColor or WHITE))] * 2
        if model.secs != 0 or model.gain != 0:
            GLib.timeout_add(100, self.update)

    def time_changed(self, model):
        self.update()

    def player_changed(self, model):
        self.redraw_canvas()

    def update(self, wmovecount=-1, bmovecount=-1):
        if self.model.ended:
            return False
        if len(self.model.gamemodel.players) < 2:
            return not self.model.ended
        alarm_time = int(self.alarm_spin)
        if self.model.getPlayerTime(self.model.movingColor) <= alarm_time and \
            self.model.gamemodel.players[self.model.movingColor].__type__ == LOCAL and \
            self.model.gamemodel.status in UNFINISHED_STATES and \
            self.model.secs != 0 and \
            self.model.gamemodel.endstatus not in (DRAW, WHITEWON, BLACKWON, UNKNOWN_STATE) and \
                not self.short_on_time[self.model.movingColor]:
            self.short_on_time[self.model.movingColor] = True
            preferencesDialog.SoundTab.playAction("shortOnTime")

        if self.model.paused and wmovecount == -1 and bmovecount == -1:
            return not self.model.ended
        white_time = formatTime(self.model.getPlayerTime(WHITE, wmovecount))
        black_time = formatTime(self.model.getPlayerTime(BLACK, bmovecount))
        if self.formatedCache != [white_time, black_time]:
            self.formatedCache = [white_time, black_time]
            self.redraw_canvas()
        return not self.model.ended



import os

from os import listdir
from os.path import isdir, isfile, splitext

from gi.repository import Gtk
import cairo

from pychess.Utils.const import *
from pychess.Utils.Piece import Piece
from pychess.gfx import Pieces
from pychess.System.prefix import addDataPrefix, getUserDataPrefix


SQUARE = 39

PIECES = ((Piece(WHITE, KING), Piece(WHITE, QUEEN), Piece(WHITE, ROOK), None),
          (Piece(WHITE, KNIGHT), Piece(WHITE, BISHOP), None, Piece(BLACK, PAWN)),
          (Piece(WHITE, PAWN), None, Piece(BLACK, BISHOP), Piece(BLACK, KNIGHT)),
          (None, Piece(BLACK, ROOK), Piece(BLACK, QUEEN), Piece(BLACK, KING)))

themes = []

pieces = addDataPrefix("pieces")
themes += [d.capitalize() for d in listdir(pieces) if isdir(os.path.join(pieces,d)) and d != 'ttf']

ttf = addDataPrefix("pieces/ttf")
themes += ['ttf-' + splitext(d)[0].capitalize() for d in listdir(ttf) if splitext(d)[1] == '.ttf']
themes.sort()

for theme in themes:
    pngfile = "%s/%s.png" % (pieces, theme)
    print('Creating %s' % pngfile)

    Pieces.set_piece_theme(theme)

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, SQUARE*4, SQUARE*4)

    context = cairo.Context(surface)
    context.set_source_rgb(0.5, 0.5, 0.5)

    for x in range(4):
        for y in range(4):
            if (x+y) % 2 == 1:
                context.rectangle(x*SQUARE, y*SQUARE, SQUARE, SQUARE)
    context.fill()

    context.rectangle(0, 0, 4*SQUARE, 4*SQUARE)
    context.stroke()

    context.set_source_rgb(0, 0, 0)
    for y, row in enumerate(PIECES):
        for x, piece in enumerate(row):
            if piece is not None:
                Pieces.drawPiece(piece, context, x*SQUARE, (3-y)*SQUARE, SQUARE)

    surface.write_to_png(pngfile)



# -*- coding: UTF-8 -*-


NAME = "PyChess"

# Player types
LOCAL, ARTIFICIAL, REMOTE = range(3)

# Engine strengths
EASY, INTERMEDIATE, EXPERT = range(3)

# Tools
TOOL_NONE, TOOL_CHESSDB, TOOL_SCOUTFISH = range(3)

# Player colors
WHITE, BLACK = range(2)

# Game states
WAITING_TO_START, PAUSED, RUNNING, DRAW, WHITEWON, BLACKWON, KILLED, \
    ADJOURNED, ABORTED, UNKNOWN_STATE, ICC_ABORTED, CANCELLED = range(12)
reprResult = ["*", "*", "*", "1/2-1/2", "1-0", "0-1", "*", "*", "*", "*", "*", "*"]

UNDOABLE_STATES = (DRAW, WHITEWON, BLACKWON)
UNFINISHED_STATES = (WAITING_TO_START, PAUSED, RUNNING, UNKNOWN_STATE)

# Chess variants
NORMALCHESS, CORNERCHESS, SHUFFLECHESS, FISCHERRANDOMCHESS, RANDOMCHESS, \
    ASYMMETRICRANDOMCHESS, UPSIDEDOWNCHESS, PAWNSPUSHEDCHESS, PAWNSPASSEDCHESS, \
    THEBANCHESS, PAWNODDSCHESS, KNIGHTODDSCHESS, ROOKODDSCHESS, QUEENODDSCHESS, \
    BLINDFOLDCHESS, HIDDENPAWNSCHESS, HIDDENPIECESCHESS, ALLWHITECHESS, \
    ATOMICCHESS, BUGHOUSECHESS, CRAZYHOUSECHESS, LOSERSCHESS, SUICIDECHESS, GIVEAWAYCHESS, \
    WILDCASTLECHESS, WILDCASTLESHUFFLECHESS, KINGOFTHEHILLCHESS, THREECHECKCHESS, HORDECHESS, \
    RACINGKINGSCHESS, ASEANCHESS, MAKRUKCHESS, SITTUYINCHESS, CAMBODIANCHESS, AIWOKCHESS, \
    EUROSHOGICHESS, SETUPCHESS, PLACEMENTCHESS, SCHESS, LIGHTBRIGADECHESS = range(40)

ASEAN_VARIANTS = (ASEANCHESS, MAKRUKCHESS, CAMBODIANCHESS, AIWOKCHESS,
                  SITTUYINCHESS)
DROP_VARIANTS = (BUGHOUSECHESS, CRAZYHOUSECHESS, EUROSHOGICHESS, SITTUYINCHESS,
                 SETUPCHESS, PLACEMENTCHESS, SCHESS)
UNSUPPORTED = (BUGHOUSECHESS, AIWOKCHESS, EUROSHOGICHESS, SETUPCHESS)

# Chess variant groups
VARIANTS_BLINDFOLD, VARIANTS_ODDS, VARIANTS_SHUFFLE, VARIANTS_OTHER, \
    VARIANTS_OTHER_NONSTANDARD, VARIANTS_ASEAN = range(6)

# Action errors
ACTION_ERROR_NOT_OUT_OF_TIME, \
    ACTION_ERROR_CLOCK_NOT_STARTED, ACTION_ERROR_SWITCH_UNDERWAY, \
    ACTION_ERROR_CLOCK_NOT_PAUSED, ACTION_ERROR_TOO_LARGE_UNDO, \
    ACTION_ERROR_NONE_TO_ACCEPT, ACTION_ERROR_NONE_TO_WITHDRAW, \
    ACTION_ERROR_NONE_TO_DECLINE, = range(8)

# Game state reasons
ABORTED_ADJUDICATION, ABORTED_AGREEMENT, ABORTED_COURTESY, ABORTED_EARLY, \
    ABORTED_SERVER_SHUTDOWN, ADJOURNED_COURTESY, ABORTED_DISCONNECTION, \
    ADJOURNED_AGREEMENT, ADJOURNED_LOST_CONNECTION, ADJOURNED_SERVER_SHUTDOWN, \
    ADJOURNED_COURTESY_WHITE, ADJOURNED_COURTESY_BLACK, \
    ADJOURNED_LOST_CONNECTION_WHITE, ADJOURNED_LOST_CONNECTION_BLACK, \
    DRAW_50MOVES, DRAW_ADJUDICATION, DRAW_AGREE, DRAW_CALLFLAG, DRAW_INSUFFICIENT, \
    DRAW_EQUALMATERIAL, DRAW_LENGTH, DRAW_REPETITION, DRAW_STALEMATE, DRAW_KINGSINEIGHTROW, \
    DRAW_BLACKINSUFFICIENTANDWHITETIME, DRAW_WHITEINSUFFICIENTANDBLACKTIME, \
    WON_ADJUDICATION, WON_CALLFLAG, WON_DISCONNECTION, WON_MATE, WON_RESIGN, \
    WON_LESSMATERIAL, WON_NOMATERIAL, WON_KINGEXPLODE, WON_KINGINCENTER, \
    WON_THREECHECK, WON_KINGINEIGHTROW, WON_WIPEOUT, PRACTICE_GOAL_REACHED, \
    WHITE_ENGINE_DIED, BLACK_ENGINE_DIED, DISCONNECTED, UNKNOWN_REASON = range(43)

ABORTED_REASONS = (ABORTED_ADJUDICATION, ABORTED_AGREEMENT, ABORTED_COURTESY,
                   ABORTED_EARLY, ABORTED_SERVER_SHUTDOWN, ABORTED_DISCONNECTION)

ADJOURNED_REASONS = (ADJOURNED_COURTESY, ADJOURNED_AGREEMENT, ADJOURNED_LOST_CONNECTION,
                     ADJOURNED_SERVER_SHUTDOWN, ADJOURNED_COURTESY_WHITE, ADJOURNED_COURTESY_BLACK,
                     ADJOURNED_LOST_CONNECTION_WHITE, ADJOURNED_LOST_CONNECTION_BLACK)

UNDOABLE_REASONS = (DRAW_50MOVES, DRAW_INSUFFICIENT, DRAW_EQUALMATERIAL,
                    DRAW_REPETITION, DRAW_STALEMATE, DRAW_AGREE, DRAW_CALLFLAG,
                    DRAW_KINGSINEIGHTROW, DRAW_BLACKINSUFFICIENTANDWHITETIME,
                    DRAW_WHITEINSUFFICIENTANDBLACKTIME,
                    WON_MATE, WON_NOMATERIAL, WON_CALLFLAG, WON_RESIGN,
                    WON_LESSMATERIAL, WON_KINGEXPLODE, WON_KINGINCENTER, WON_THREECHECK,
                    WON_KINGINEIGHTROW, WON_WIPEOUT)

ADJUDICATION_REASONS = (DRAW_ADJUDICATION, WON_ADJUDICATION)

DEATH_REASONS = (WHITE_ENGINE_DIED, BLACK_ENGINE_DIED, DISCONNECTED)

CALLFLAG_REASONS = (DRAW_CALLFLAG, WON_CALLFLAG)

# Player actions
RESIGNATION = "resignation"
FLAG_CALL = "flag call"
DRAW_OFFER = "draw offer"
ABORT_OFFER = "abort offer"
ADJOURN_OFFER = "adjourn offer"
PAUSE_OFFER = "pause offer"
RESUME_OFFER = "resume offer"
SWITCH_OFFER = "switch offer"
TAKEBACK_OFFER = "takeback offer"
MATCH_OFFER = "match offer"
HURRY_ACTION = "hurry action"
CHAT_ACTION = "chat action"

ACTIONS = (RESIGNATION, FLAG_CALL, DRAW_OFFER, ABORT_OFFER, ADJOURN_OFFER,
           PAUSE_OFFER, RESUME_OFFER, SWITCH_OFFER, TAKEBACK_OFFER,
           MATCH_OFFER, HURRY_ACTION, CHAT_ACTION)
OFFERS = (DRAW_OFFER, ABORT_OFFER, ADJOURN_OFFER, PAUSE_OFFER,
          RESUME_OFFER, SWITCH_OFFER, TAKEBACK_OFFER, MATCH_OFFER)
INGAME_ACTIONS = (RESIGNATION, FLAG_CALL, DRAW_OFFER, ABORT_OFFER,
                  ADJOURN_OFFER, PAUSE_OFFER, SWITCH_OFFER, HURRY_ACTION)

# A few nice to have boards
FEN_EMPTY = "4k3/8/8/8/8/8/8/4K3 w - - 0 1"
FEN_START = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"
FEN_START_960 = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w HAha - 0 1"

# Search values

hashfALPHA, hashfBETA, hashfEXACT, hashfBAD = range(4)

# Engine modes
NORMAL, ANALYZING, INVERSE_ANALYZING = range(3)

# Piece types

# BPAWN is a pawn that moves in the opposite direction
EMPTY, PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING, HAWK, ELEPHANT, BPAWN, \
    ASEAN_WBISHOP, ASEAN_BBISHOP, ASEAN_QUEEN = range(13)

# Is sliding piece
sliders = [False, False, False, True, True, True, False,
           True, True, False, False, False, False]

# Piece signs
reprSign = ["", "P", "N", "B", "R", "Q", "K", "H", "E"]
reprSignMakruk = ["", "P", "N", "S", "R", "M", "K"]
reprSignSittuyin = ["", "P", "N", "S", "R", "F", "K"]
chr2Sign = {"k": KING,
            "q": QUEEN,
            "r": ROOK,
            "b": BISHOP,
            "n": KNIGHT,
            "p": PAWN,
            "m": QUEEN,
            "s": BISHOP,
            "f": QUEEN,
            "h": HAWK,
            "e": ELEPHANT}
chrU2Sign = {"K": KING,
             "Q": QUEEN,
             "R": ROOK,
             "B": BISHOP,
             "N": KNIGHT,
             "P": PAWN,
             "M": QUEEN,
             "S": BISHOP,
             "F": QUEEN,
             "H": HAWK,
             "E": ELEPHANT}

# Move values
NORMAL_MOVE, QUEEN_CASTLE, KING_CASTLE, ENPASSANT, \
    KNIGHT_PROMOTION, BISHOP_PROMOTION, ROOK_PROMOTION, QUEEN_PROMOTION, KING_PROMOTION, \
    HAWK_PROMOTION, ELEPHANT_PROMOTION, \
    HAWK_GATE, ELEPHANT_GATE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK, DROP = range(16)

PROMOTIONS = (KING_PROMOTION, QUEEN_PROMOTION, ROOK_PROMOTION,
              BISHOP_PROMOTION, KNIGHT_PROMOTION, HAWK_PROMOTION, ELEPHANT_PROMOTION)

GATINGS = (HAWK_GATE, ELEPHANT_GATE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)

# Algebraic notation types: Short, Long, Figure and Simpe
SAN, LAN, FAN, AN = range(4)
# Castling notation types: e.g., O-O, e1g1, e1h1
CASTLE_SAN, CASTLE_KK, CASTLE_KR = range(3)

FAN_PIECES = [
    ["", "♙", "♘", "♗", "♖", "♕", "♔", "H", "E", ""],
    ["", "♟", "♞", "♝", "♜", "♛", "♚", "h", "e", ""]
]

# Castling values
W_OO, W_OOO, B_OO, B_OOO = [2**i for i in range(4)]
CAS_FLAGS = ((W_OOO, W_OO), (B_OOO, B_OO))
W_CASTLED, B_CASTLED = [2**i for i in range(2)]

NB_OF_CASES = 64

# Cords types
A1, B1, C1, D1, E1, F1, G1, H1,\
    A2, B2, C2, D2, E2, F2, G2, H2,\
    A3, B3, C3, D3, E3, F3, G3, H3,\
    A4, B4, C4, D4, E4, F4, G4, H4,\
    A5, B5, C5, D5, E5, F5, G5, H5,\
    A6, B6, C6, D6, E6, F6, G6, H6,\
    A7, B7, C7, D7, E7, F7, G7, H7,\
    A8, B8, C8, D8, E8, F8, G8, H8 = range(64)

reprCord = [
    "a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1",
    "a2", "b2", "c2", "d2", "e2", "f2", "g2", "h2",
    "a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3",
    "a4", "b4", "c4", "d4", "e4", "f4", "g4", "h4",
    "a5", "b5", "c5", "d5", "e5", "f5", "g5", "h5",
    "a6", "b6", "c6", "d6", "e6", "f6", "g6", "h6",
    "a7", "b7", "c7", "d7", "e7", "f7", "g7", "h7",
    "a8", "b8", "c8", "d8", "e8", "f8", "g8", "h8"
]

reprFile = ["a", "b", "c", "d", "e", "f", "g", "h"]
reprRank = ["1", "2", "3", "4", "5", "6", "7", "8"]

# dictionary for giving the cord linked to the flattened index of a coordinate
cordDic = {}
for cord, name in enumerate(reprCord):
    cordDic[name] = cord

# User interface

# pagination in database game list
FIRST_PAGE = 0
PREV_PAGE = -1
NEXT_PAGE = 1

# Hint modes
OPENING, ENDGAME, HINT, SPY = ["opening", "endgame", "hint", "spy"]

# Sound settings
SOUND_MUTE, SOUND_BEEP, SOUND_SELECT, SOUND_URI = range(4)
COUNT_OF_SOUNDS = 13

# Brush types. Send piece object for Piece brush
CLEAR, ENPAS = range(2)

# Main menu items
GAME_MENU_ITEMS = ("save_game1", "save_game_as1", "share_game",
                   "export_position1", "analyze_game1", "properties1",
                   "close1")
ACTION_MENU_ITEMS = ("abort", "adjourn", "draw", "pause1", "resume1", "undo1",
                     "call_flag", "resign", "ask_to_move")
VIEW_MENU_ITEMS = ("rotate_board1", "show_sidepanels", "hint_mode", "spy_mode")
EDIT_MENU_ITEMS = ("copy_pgn", "copy_fen", )
MENU_ITEMS = GAME_MENU_ITEMS + ACTION_MENU_ITEMS + VIEW_MENU_ITEMS + EDIT_MENU_ITEMS

# Column name
COLUMN_ROW_RESET = "column_row_reset"

# Gtk icon name
GTK_ICON_VIEW_REFRESH = "view-refresh"

# Learn categories
LECTURE, LESSON, PUZZLE, ENDGAME = range(4)



# An opening book needs to be built as follows:
#   - install "pgn-extract" and "polyglot"
#   - collect many relevant games into a single PGN file "input.pgn"
#   - remove the variant games (960, atomic...) from the file
#   - solve the errors given by: pgn-extract -s -r input.pgn
#   - create a file as "filter.txt":
#           WhiteElo >= "1800"
#           BlackElo >= "1800"
#   - extract the best games: pgn-extract -tfilter.txt --notags --nocomments --nonags --novars -bl20 --plylimit 26 -s -owip.pgn input.pgn
#   - create the opening book: polyglot make-book -min-game 10 -pgn wip.pgn -bin pychess_book.bin
#   - merge the books (if needed): polyglot merge-book -in1 book1.bin -in2 book2.bin -out book.bin
#
# The opening book does not contain the names. They are stored in the separate file "eco.db" by
# running the current script. If a name refers to a position that is not part of the opening book,
# it cannot be displayed. If it refers to a shared position, the name is selected according to some
# priority rules. The source ECO file must also be sorted by ECO (at least) to be able to load the
# other languages.
#
# The current book supports 99.5% of the ECO names written in English.

import sys
import os
import sqlite3

from pychess.Savers.pgn import load
from pychess.System.protoopen import protoopen
from pychess.System.prefix import addDataPrefix
from pychess.Utils.eco import ECO_MAIN_LANG, ECO_LANGS
from pychess.Variants.fischerandom import FischerandomBoard


path = os.path.join(addDataPrefix("eco.db"))
conn = sqlite3.connect(path)

if __name__ == '__main__':
    print("Creating the database")
    c = conn.cursor()
    c.execute("drop table if exists openings")
    c.execute("create table openings (hash text, hkey integer, mainline integer, endline integer, eco text, lang text, opening text, variation text, fen text)")
    c.execute("create index if not exists openings_index on openings (hkey)")

    def feed(pgnfile, lang):
        # Check the existence of the file
        if not os.path.isfile(pgnfile):
            return

        # Load the ECO file first
        print('  - Parsing')
        cf = load(protoopen(pgnfile))
        cf.limit = 5000
        cf.init_tag_database()
        records, plys = cf.get_records()

        # Cache the content
        entries = []
        plyMax = 0
        old_eco = ""
        for rec in records:
            model = cf.loadToModel(rec)
            eco = '' if rec['ECO'] is None else rec['ECO']
            entry = {'h': [],                                               # Hashes
                     'f': '',                                               # Final hash of the line
                     'n': [],                                               # FENs
                     'm': old_eco != eco,                                   # Main line = shortest sequence of moves for the ECO code. The 'EN' ECO file is specially crafted
                     'e': eco,                                              # ECO
                     'o': '' if rec['White'] is None else rec['White'],     # Opening
                     'v': '' if rec['Black'] is None else rec['Black'],     # Variation
                     'p': len(model.moves)}                                 # Number of plies
            plyMax = max(plyMax, entry['p'])

            # No move means that we are translating the name of the ECO code, so we need to find all the related positions from another language
            if entry['p'] == 0:
                if lang == ECO_MAIN_LANG:
                    continue
                c.execute("select hash, endline, fen from openings where eco=? and lang=? and mainline=1", (eco, ECO_MAIN_LANG))
                rows = c.fetchall()
                for row in rows:
                    entry['h'].append(row[0])
                    if (row[1] == int(True)):
                        entry['f'] = row[0]
                    entry['n'].append(row[2])
            else:
                # Find the Polyglot hash for each position of the opening
                for i in range(entry['p']):
                    nextboard = model.getBoardAtPly(i, 0).board.next
                    h = hex(nextboard.hash)[2:]
                    entry['h'].append(h)
                    entry['f'] = h
                    entry['n'].append(nextboard.asFen())
            entries.append(entry)
            old_eco = entry['e']
        print('  - Max ply : %d' % plyMax)

        # Process all the data in reverse order
        for depth in reversed(range(plyMax + 1)):
            sys.stdout.write("\r  - Loading into the database (%d remaining)  " % depth)
            sys.stdout.flush()
            for i in reversed(range(len(entries))):  # Long lines are overwritten by short lines
                entry = entries[i]
                if entry['p'] != depth:
                    continue
                for i in range(len(entry['h'])):
                    h = entry['h'][i]
                    hkey = int(h[-2:], 16)
                    c.execute("select endline from openings where hash=? and hkey=? and lang=?", (h, hkey, lang))
                    r = c.fetchone()
                    if r is not None and r[0] == int(True):
                        continue
                    c.execute("delete from openings where hash=? and hkey=? and lang=?", (h, hkey, lang))
                    c.execute("insert into openings (hash, hkey, mainline, endline, eco, lang, opening, variation, fen) values (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                              (h, hkey, int(entry['m']), int(h == entry['f']), entry['e'], lang, entry['o'], entry['v'], entry['n'][i]))
        conn.commit()
        print('\n  - Processed %d openings' % len(entries))

    # Several eco lists contain only eco+name pairs
    # We use the base ECO line positions from EN/eco.pgn
    # English is first in ECO_LANGS for that reason
    for lang in ECO_LANGS:
        print("Processing %s" % lang.upper())
        feed("lang/%s/eco.pgn" % lang, lang)

    # Start positions for Chess960
    print("Processing Chess960")
    chess960 = FischerandomBoard()
    for i in range(960):
        c.execute("insert into openings (mainline, endline, eco, lang, opening, fen) values (?, '1', '960', ?, ?, ?)",
                  ('1' if i == 518 else '0',
                   ECO_MAIN_LANG,
                   'Chess%.3d' % (i + 1),
                   chess960.getFrcFen(i + 1)))
    conn.close()



from gi.repository import Gtk, Gdk, GdkPixbuf

from pychess.ic.FICSObjects import FICSSoughtMatch, FICSChallenge, \
    get_seek_tooltip_text, get_challenge_tooltip_text
from pychess.ic import TYPE_BLITZ, TYPE_LIGHTNING, TYPE_STANDARD, RATING_TYPES, \
    TYPE_BULLET, TYPE_ONE_MINUTE, TYPE_THREE_MINUTE, TYPE_FIVE_MINUTE, \
    TYPE_FIFTEEN_MINUTE, TYPE_FORTYFIVE_MINUTE, \
    get_infobarmessage_content
from pychess.perspectives.fics.ParrentListSection import ParrentListSection, cmp, \
    SEPARATOR, ACCEPT, ASSESS, FOLLOW, CHAT, CHALLENGE, FINGER, ARCHIVED
from pychess.Utils.IconLoader import get_pixbuf
from pychess.System import conf, uistuff
from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix
from pychess.widgets import mainwindow
from pychess.widgets.preferencesDialog import SoundTab
from pychess.widgets.InfoBar import InfoBarMessage, InfoBarMessageButton

__title__ = _("Seeks / Challenges")

__icon__ = addDataPrefix("glade/manseek.svg")

__desc__ = _("Handle seeks and challenges")


class Sidepanel(ParrentListSection):

    def load(self, widgets, connection, lounge):
        self.widgets = widgets
        self.connection = connection
        self.lounge = lounge
        self.infobar = lounge.infobar

        __widget__ = lounge.seek_list

        self.messages = {}
        self.seeks = {}
        self.challenges = {}
        self.seekPix = get_pixbuf("glade/seek.png")
        self.chaPix = get_pixbuf("glade/challenge.png")
        self.manSeekPix = get_pixbuf("glade/manseek.png")

        self.widgets["seekExpander"].set_vexpand(False)

        self.tv = self.widgets["seektreeview"]
        self.store = Gtk.ListStore(FICSSoughtMatch, GdkPixbuf.Pixbuf,
                                   GdkPixbuf.Pixbuf, str, int, str, str, str,
                                   int, Gdk.RGBA, str)

        self.seek_filter = self.store.filter_new()
        self.seek_filter.set_visible_func(self.seek_filter_func)

        self.filter_toggles = {}
        self.filter_buttons = ("standard_toggle1", "blitz_toggle1", "lightning_toggle1", "variant_toggle1", "computer_toggle1")
        for widget in self.filter_buttons:
            uistuff.keep(self.widgets[widget], widget)
            self.widgets[widget].connect("toggled", self.on_filter_button_toggled)
            initial = conf.get(widget)
            self.filter_toggles[widget] = initial
            self.widgets[widget].set_active(initial)

        self.model = self.seek_filter.sort_new_with_model()
        self.tv.set_model(self.model)

        self.tv.set_model(self.model)
        self.addColumns(self.tv,
                        "FICSSoughtMatch",
                        "",
                        "",
                        _("Name"),
                        _("Rating"),
                        _("Rated"),
                        _("Type"),
                        _("Clock"),
                        "gametime",
                        "textcolor",
                        "tooltip",
                        hide=[0, 8, 9, 10],
                        pix=[1, 2])
        self.tv.set_search_column(3)
        self.tv.set_tooltip_column(10, )
        for i in range(0, 2):
            self.tv.get_model().set_sort_func(i, self.pixCompareFunction,
                                              i + 1)
        for i in range(2, 8):
            self.tv.get_model().set_sort_func(i, self.compareFunction, i)
        try:
            self.tv.set_search_position_func(self.lowLeftSearchPosFunc, None)
        except AttributeError:
            # Unknow signal name is raised by gtk < 2.10
            pass
        for num in range(2, 7):
            column = self.tv.get_column(num)
            for cellrenderer in column.get_cells():
                column.add_attribute(cellrenderer, "foreground_rgba", 9)
        self.selection = self.tv.get_selection()
        self.lastSeekSelected = None
        self.selection.set_select_function(self.selectFunction, True)
        self.selection.connect("changed", self.onSelectionChanged)
        self.widgets["clearSeeksButton"].connect("clicked",
                                                 self.onClearSeeksClicked)
        self.widgets["acceptButton"].connect("clicked", self.on_accept)
        self.widgets["declineButton"].connect("clicked", self.onDeclineClicked)
        self.tv.connect("row-activated", self.row_activated)
        self.tv.connect('button-press-event', self.button_press_event)

        self.connection.seeks.connect("FICSSeekCreated", self.onAddSeek)
        self.connection.seeks.connect("FICSSeekRemoved", self.onRemoveSeek)
        self.connection.challenges.connect("FICSChallengeIssued",
                                           self.onChallengeAdd)
        self.connection.challenges.connect("FICSChallengeRemoved",
                                           self.onChallengeRemove)
        self.connection.glm.connect("our-seeks-removed",
                                    self.our_seeks_removed)
        self.connection.glm.connect("assessReceived", self.onAssessReceived)
        self.connection.bm.connect("playGameCreated", self.onPlayingGame)
        self.connection.bm.connect("curGameEnded", self.onCurGameEnded)

        def get_sort_order(modelsort):
            identity, order = modelsort.get_sort_column_id()
            if identity is None or identity < 0:
                identity = 0
            else:
                identity += 1
            if order == Gtk.SortType.DESCENDING:
                identity = -1 * identity
            return identity

        def set_sort_order(modelsort, value):
            if value != 0:
                order = Gtk.SortType.ASCENDING if value > 0 else Gtk.SortType.DESCENDING
                modelsort.set_sort_column_id(abs(value) - 1, order)

        uistuff.keep(self.model, "seektreeview_sort_order_col", get_sort_order,
                     lambda modelsort, value: set_sort_order(modelsort, value))

        self.createLocalMenu((ACCEPT, ASSESS, CHALLENGE, CHAT, FOLLOW, SEPARATOR, FINGER, ARCHIVED))
        self.assess_sent = False

        return __widget__

    def seek_filter_func(self, model, iter, data):
        sought_match = model[iter][0]
        is_computer = sought_match.player.isComputer()
        is_standard = sought_match.game_type.rating_type in (TYPE_STANDARD, TYPE_FIFTEEN_MINUTE, TYPE_FORTYFIVE_MINUTE) and not is_computer
        is_blitz = sought_match.game_type.rating_type in (TYPE_BLITZ, TYPE_THREE_MINUTE, TYPE_FIVE_MINUTE) and not is_computer
        is_lightning = sought_match.game_type.rating_type in (TYPE_LIGHTNING, TYPE_BULLET, TYPE_ONE_MINUTE) and not is_computer
        is_variant = sought_match.game_type.rating_type in RATING_TYPES[9:] and not is_computer
        return (
            self.filter_toggles["computer_toggle1"] and is_computer) or (
            self.filter_toggles["standard_toggle1"] and is_standard) or (
            self.filter_toggles["blitz_toggle1"] and is_blitz) or (
            self.filter_toggles["lightning_toggle1"] and is_lightning) or (
            self.filter_toggles["variant_toggle1"] and is_variant)

    def on_filter_button_toggled(self, widget):
        for button in self.filter_buttons:
            self.filter_toggles[button] = self.widgets[button].get_active()
        self.seek_filter.refilter()

    def onAssessReceived(self, glm, assess):
        if self.assess_sent:
            self.assess_sent = False
            dialog = Gtk.MessageDialog(mainwindow(), type=Gtk.MessageType.INFO,
                                       buttons=Gtk.ButtonsType.OK)
            dialog.set_title(_("Assess"))
            dialog.set_markup(_("Effect on ratings by the possible outcomes"))
            grid = Gtk.Grid()
            grid.set_column_homogeneous(True)
            grid.set_row_spacing(12)
            grid.set_row_spacing(12)
            name0 = Gtk.Label()
            name0.set_markup("<b>%s</b>" % assess["names"][0])
            name1 = Gtk.Label()
            name1.set_markup("<b>%s</b>" % assess["names"][1])
            grid.attach(Gtk.Label(label=""), 0, 0, 1, 1)
            grid.attach(name0, 1, 0, 1, 1)
            grid.attach(name1, 2, 0, 1, 1)
            grid.attach(Gtk.Label(assess["type"]), 0, 1, 1, 1)
            grid.attach(Gtk.Label(assess["oldRD"][0]), 1, 1, 1, 1)
            grid.attach(Gtk.Label(assess["oldRD"][1]), 2, 1, 1, 1)
            grid.attach(Gtk.Label(_("Win:")), 0, 2, 1, 1)
            grid.attach(Gtk.Label(assess["win"][0]), 1, 2, 1, 1)
            grid.attach(Gtk.Label(assess["win"][1]), 2, 2, 1, 1)
            grid.attach(Gtk.Label(_("Draw:")), 0, 3, 1, 1)
            grid.attach(Gtk.Label(assess["draw"][0]), 1, 3, 1, 1)
            grid.attach(Gtk.Label(assess["draw"][1]), 2, 3, 1, 1)
            grid.attach(Gtk.Label(_("Loss:")), 0, 4, 1, 1)
            grid.attach(Gtk.Label(assess["loss"][0]), 1, 4, 1, 1)
            grid.attach(Gtk.Label(assess["loss"][1]), 2, 4, 1, 1)
            grid.attach(Gtk.Label(_("New RD:")), 0, 5, 1, 1)
            grid.attach(Gtk.Label(assess["newRD"][0]), 1, 5, 1, 1)
            grid.attach(Gtk.Label(assess["newRD"][1]), 2, 5, 1, 1)
            grid.show_all()
            dialog.get_message_area().add(grid)
            dialog.run()
            dialog.destroy()

    def getSelectedPlayer(self):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is not None:
            sought = model.get_value(sel_iter, 0)
            return sought.player

    def textcolor_normal(self):
        style_ctxt = self.tv.get_style_context()
        return style_ctxt.get_color(Gtk.StateFlags.NORMAL)

    def textcolor_selected(self):
        style_ctxt = self.tv.get_style_context()
        return style_ctxt.get_color(Gtk.StateFlags.INSENSITIVE)

    def selectFunction(self, selection, model, path, is_selected, data):
        if model[path][9] == self.textcolor_selected():
            return False
        else:
            return True

    def __isAChallengeOrOurSeek(self, row):
        sought = row[0]
        textcolor = row[9]
        red0, green0, blue0 = textcolor.red, textcolor.green, textcolor.blue
        selected = self.textcolor_selected()
        red1, green1, blue1 = selected.red, selected.green, selected.blue
        if (isinstance(sought, FICSChallenge)) or (red0 == red1 and green0 == green1 and
                                                   blue0 == blue1):
            return True
        else:
            return False

    def compareFunction(self, model, iter0, iter1, column):
        row0 = list(model[model.get_path(iter0)])
        row1 = list(model[model.get_path(iter1)])
        is_ascending = True if self.tv.get_column(column - 1).get_sort_order() is \
            Gtk.SortType.ASCENDING else False
        if self.__isAChallengeOrOurSeek(
                row0) and not self.__isAChallengeOrOurSeek(row1):
            if is_ascending:
                return -1
            else:
                return 1
        elif self.__isAChallengeOrOurSeek(
                row1) and not self.__isAChallengeOrOurSeek(row0):
            if is_ascending:
                return 1
            else:
                return -1
        elif column == 7:
            return self.timeCompareFunction(model, iter0, iter1, column)
        else:
            value0 = row0[column]
            value0 = value0.lower() if isinstance(value0, str) else value0
            value1 = row1[column]
            value1 = value1.lower() if isinstance(value1, str) else value1
            return cmp(value0, value1)

    def __updateActiveSeeksLabel(self):
        count = len(self.seeks) + len(self.challenges)
        self.widgets["activeSeeksLabel"].set_text(_("Active seeks: %d") %
                                                  count)

    def onAddSeek(self, seeks, seek):
        log.debug("%s" % seek,
                  extra={"task": (self.connection.username, "onAddSeek")})
        pix = self.seekPix if seek.automatic else self.manSeekPix
        textcolor = self.textcolor_selected() if seek.player.name == self.connection.getUsername() \
            else self.textcolor_normal()
        seek_ = [seek, seek.player.getIcon(gametype=seek.game_type), pix,
                 seek.player.name + seek.player.display_titles(),
                 seek.player_rating, seek.display_rated,
                 seek.game_type.display_text, seek.display_timecontrol,
                 seek.sortable_time, textcolor, get_seek_tooltip_text(seek)]

        if textcolor == self.textcolor_selected():
            txi = self.store.prepend(seek_)
            self.tv.scroll_to_cell(self.store.get_path(txi))
            self.widgets["clearSeeksButton"].set_sensitive(True)
        else:
            txi = self.store.append(seek_)
        self.seeks[hash(seek)] = txi
        self.__updateActiveSeeksLabel()

    def onRemoveSeek(self, seeks, seek):
        log.debug("%s" % seek,
                  extra={"task": (self.connection.username, "onRemoveSeek")})
        try:
            treeiter = self.seeks[hash(seek)]
        except KeyError:
            # We ignore removes we haven't added, as it seems fics sends a
            # lot of removes for games it has never told us about
            return
        if self.store.iter_is_valid(treeiter):
            self.store.remove(treeiter)
        del self.seeks[hash(seek)]
        self.__updateActiveSeeksLabel()

    def onChallengeAdd(self, challenges, challenge):
        log.debug("%s" % challenge,
                  extra={"task": (self.connection.username, "onChallengeAdd")})
        SoundTab.playAction("aPlayerChecks")

        # TODO: differentiate between challenges and manual-seek-accepts
        # (wait until seeks are comparable FICSSeek objects to do this)
        # Related: http://code.google.com/p/pychess/issues/detail?id=206
        if challenge.adjourned:
            text = _(" would like to resume your adjourned <b>%(time)s</b> " +
                     "<b>%(gametype)s</b> game.") % \
                {"time": challenge.display_timecontrol,
                 "gametype": challenge.game_type.display_text}
        else:
            text = _(" challenges you to a <b>%(time)s</b> %(rated)s <b>%(gametype)s</b> game") \
                % {"time": challenge.display_timecontrol,
                   "rated": challenge.display_rated.lower(),
                   "gametype": challenge.game_type.display_text}
            if challenge.color:
                text += _(" where <b>%(player)s</b> plays <b>%(color)s</b>.") \
                    % {"player": challenge.player.name,
                       "color": _("white") if challenge.color == "white" else _("black")}
            else:
                text += "."
        content = get_infobarmessage_content(challenge.player,
                                             text,
                                             gametype=challenge.game_type)

        def callback(infobar, response, message):
            if response == Gtk.ResponseType.ACCEPT:
                self.connection.om.acceptIndex(challenge.index)
            elif response == Gtk.ResponseType.NO:
                self.connection.om.declineIndex(challenge.index)
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.QUESTION, content, callback)
        message.add_button(InfoBarMessageButton(
            _("Accept"), Gtk.ResponseType.ACCEPT))
        message.add_button(InfoBarMessageButton(
            _("Decline"), Gtk.ResponseType.NO))
        message.add_button(InfoBarMessageButton(Gtk.STOCK_CLOSE,
                                                Gtk.ResponseType.CANCEL))
        self.messages[hash(challenge)] = message
        self.infobar.push_message(message)

        txi = self.store.prepend(
            [challenge, challenge.player.getIcon(gametype=challenge.game_type),
             self.chaPix, challenge.player.name +
             challenge.player.display_titles(), challenge.player_rating,
             challenge.display_rated, challenge.game_type.display_text,
             challenge.display_timecontrol, challenge.sortable_time,
             self.textcolor_normal(), get_challenge_tooltip_text(challenge)])
        self.challenges[hash(challenge)] = txi
        self.__updateActiveSeeksLabel()
        self.widgets["seektreeview"].scroll_to_cell(self.store.get_path(txi))

    def onChallengeRemove(self, challenges, challenge):
        log.debug(
            "%s" % challenge,
            extra={"task": (self.connection.username, "onChallengeRemove")})
        try:
            txi = self.challenges[hash(challenge)]
        except KeyError:
            pass
        else:
            if self.store.iter_is_valid(txi):
                self.store.remove(txi)
            del self.challenges[hash(challenge)]

        try:
            message = self.messages[hash(challenge)]
        except KeyError:
            pass
        else:
            message.dismiss()
            del self.messages[hash(challenge)]
        self.__updateActiveSeeksLabel()

    def our_seeks_removed(self, glm):
        self.widgets["clearSeeksButton"].set_sensitive(False)

    def onDeclineClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        sought = model.get_value(sel_iter, 0)
        self.connection.om.declineIndex(sought.index)

        try:
            message = self.messages[hash(sought)]
        except KeyError:
            pass
        else:
            message.dismiss()
            del self.messages[hash(sought)]

    def onClearSeeksClicked(self, button):
        self.connection.client.run_command("unseek")
        self.widgets["clearSeeksButton"].set_sensitive(False)

    def row_activated(self, treeview, path, view_column):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        sought = model.get_value(sel_iter, 0)
        if self.lastSeekSelected is None or \
                sought.index != self.lastSeekSelected.index:
            return
        if path != model.get_path(sel_iter):
            return
        self.on_accept(None)

    def onSelectionChanged(self, selection):
        model, sel_iter = selection.get_selected()
        sought = None
        a_seek_is_selected = False
        selection_is_challenge = False
        if sel_iter is not None:
            a_seek_is_selected = True
            sought = model.get_value(sel_iter, 0)
            if isinstance(sought, FICSChallenge):
                selection_is_challenge = True

            # # select sought owner on players tab to let challenge him using right click menu
            # if sought.player in self.lounge.players_tab.players:
                # # we have to undo the iter conversion that was introduced by the filter and sort model
                # iter0 = self.lounge.players_tab.players[sought.player]["ti"]
                # filtered_model = self.lounge.players_tab.player_filter
                # is_ok, iter1 = filtered_model.convert_child_iter_to_iter(iter0)
                # sorted_model = self.lounge.players_tab.model
                # is_ok, iter2 = sorted_model.convert_child_iter_to_iter(iter1)
                # players_selection = self.lounge.players_tab.tv.get_selection()
                # players_selection.select_iter(iter2)
                # self.lounge.players_tab.tv.scroll_to_cell(sorted_model.get_path(iter2))
            # else:
                # print(sought.player, "not in self.lounge.players_tab.players")

        self.lastSeekSelected = sought
        self.widgets["acceptButton"].set_sensitive(a_seek_is_selected)
        self.widgets["declineButton"].set_sensitive(selection_is_challenge)

    def _clear_messages(self):
        for message in self.messages.values():
            message.dismiss()
        self.messages.clear()

    def onPlayingGame(self, bm, game):
        self._clear_messages()
        self.widgets["seekListContent"].set_sensitive(False)
        self.widgets["clearSeeksButton"].set_sensitive(False)
        self.__updateActiveSeeksLabel()

    def onCurGameEnded(self, bm, game):
        self.widgets["seekListContent"].set_sensitive(True)



""" Random Chess """

import random

from pychess.Utils.const import RANDOMCHESS, VARIANTS_SHUFFLE
from pychess.Utils.Board import Board


class RandomBoard(Board):
    """ :Description:
    * Randomly chosen pieces (two queens or three rooks possible)
    * Exactly one king of each color
    * Pieces placed randomly behind the pawns
    * No castling
    * Black's arrangement mirrors white's
    """

    variant = RANDOMCHESS
    __desc__ = _("FICS wild/3: http://www.freechess.org/Help/HelpFiles/wild.html\n" +
                 "* Randomly chosen pieces (two queens or three rooks possible)\n" +
                 "* Exactly one king of each color\n" +
                 "* Pieces placed randomly behind the pawns\n" +
                 "* No castling\n" +
                 "* Black's arrangement mirrors white's")
    name = _("Random")
    cecp_name = "wild/3"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.random_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def random_start(self):
        back_rank = random.sample(('r', 'n', 'b', 'q') * 16, 7)
        back_rank.append('k')
        random.shuffle(back_rank)
        fen = ''.join(back_rank)
        fen = fen + '/pppppppp/8/8/8/8/PPPPPPPP/' + fen.upper() + ' w - - 0 1'
        return fen


if __name__ == '__main__':
    Board = RandomBoard(True)
    for i in range(10):
        print(Board.random_start())



from io import StringIO
from urllib.error import URLError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from gi.repository import Gdk, Gtk

from pychess.Utils.const import NAME, UNDOABLE_STATES
from pychess.Savers import pgn
from pychess.widgets import mainwindow

URL = "https://www.chesspastebin.com/api/add/"
APIKEY = "a137d919b75c8766b082367610189358cfb1ba70"


def paste(gamemodel):
    dialog = Gtk.MessageDialog(mainwindow(), type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
    if gamemodel.status in UNDOABLE_STATES:
        text = _("The current game is over. First, please verify the properties of the game.")
    else:
        text = _("The current game is not terminated. Its export may have a limited interest.")
    text += "\n\n" + _("Should %s publicly publish your game as PGN on chesspastebin.com ?") % NAME
    dialog.set_markup(text)
    response = dialog.run()
    dialog.destroy()
    if response != Gtk.ResponseType.YES:
        return

    output = StringIO()
    text = pgn.save(output, gamemodel)
    values = {'apikey': APIKEY,
              'pgn': text,
              "name": "PyChess",
              'sandbox': 'false'}

    data = urlencode(values).encode('utf-8')
    req = Request(URL, data)
    try:
        response = urlopen(req, timeout=10)
    except URLError as err:
        if hasattr(err, 'reason'):
            print('We failed to reach the server.')
            print('Reason: ', err.reason)
        elif hasattr(err, 'code'):
            print('The server couldn\'t fulfill the request.')
            print('Error code: ', err.code)
    else:
        ID = response.read()
        link = "http://www.chesspastebin.com/?p=%s" % int(ID)
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(link, -1)
        # print(text)
        # print(clipboard.wait_for_text())
        msg_dialog = Gtk.MessageDialog(mainwindow(), type=Gtk.MessageType.INFO,
                                       buttons=Gtk.ButtonsType.OK)
        msg = _(
            "Game shared at ") + '<a href="%s">chesspastebin.com</a>' % link
        msg_dialog.set_markup(msg)
        msg_dialog.format_secondary_text(_("(Link is available on clipboard.)"))
        msg_dialog.connect("response", lambda msg_dialog, a: msg_dialog.hide())
        msg_dialog.show()



import builtins

from .const import ADJOURNED_LOST_CONNECTION_BLACK, UNKNOWN_REASON, WON_LESSMATERIAL, \
    KILLED, ADJOURNED_SERVER_SHUTDOWN, DRAW_EQUALMATERIAL, \
    ABORTED_ADJUDICATION, DISCONNECTED, BLACKWON, WHITEWON, DRAW_CALLFLAG, WON_RESIGN, DRAW, \
    WON_CALLFLAG, DRAW_WHITEINSUFFICIENTANDBLACKTIME, DRAW_LENGTH, WON_MATE, WON_KINGINEIGHTROW, \
    DRAW_BLACKINSUFFICIENTANDWHITETIME, ADJOURNED_COURTESY_WHITE, WON_THREECHECK, WON_WIPEOUT, \
    DRAW_AGREE, DRAW_INSUFFICIENT, DRAW_STALEMATE, DRAW_ADJUDICATION, UNKNOWN_STATE, \
    ADJOURNED_LOST_CONNECTION, WON_NOMATERIAL, ADJOURNED, WON_DISCONNECTION, ABORTED, \
    ADJOURNED_AGREEMENT, WON_KINGEXPLODE, WON_KINGINCENTER, ABORTED_SERVER_SHUTDOWN, \
    ABORTED_AGREEMENT, WHITE_ENGINE_DIED, BLACK_ENGINE_DIED, WON_ADJUDICATION, \
    ADJOURNED_COURTESY_BLACK, ADJOURNED_COURTESY, DRAW_REPETITION, ABORTED_COURTESY, \
    DRAW_50MOVES, ADJOURNED_LOST_CONNECTION_WHITE, ABORTED_EARLY, ABORTED_DISCONNECTION, \
    CANCELLED, PRACTICE_GOAL_REACHED, DRAW_KINGSINEIGHTROW


if '_' not in builtins.__dict__:
    builtins.__dict__['_'] = lambda s: s
    builtins.__dict__[
        'ngettext'] = lambda singular, plural, n: singular if n == 1 else plural

reprColor = [_("White"), _("Black")]

reprPiece = ["Empty", _("Pawn"), _("Knight"), _("Bishop"), _("Rook"),
             _("Queen"), _("King"), _("Hawk"), _("Elephant"), "BPawn"]

localReprSign = ["", _("P"), _("N"), _("B"), _("R"), _("Q"), _("K"), _("H"), _("E")]

reprResult_long = {
    DRAW: _("The game ended in a draw"),
    WHITEWON: _("%(white)s won the game"),
    BLACKWON: _("%(black)s won the game"),
    KILLED: _("The game has been killed"),
    ADJOURNED: _("The game has been adjourned"),
    ABORTED: _("The game has been aborted"),
    UNKNOWN_STATE: _("Unknown game state"),
    CANCELLED: _("Game cancelled"),
}

reprReason_long = {
    DRAW_INSUFFICIENT:
    _("Because neither player has sufficient material to mate"),
    DRAW_REPETITION:
    _("Because the same position was repeated three times in a row"),
    DRAW_50MOVES: _("Because the last 50 moves brought nothing new"),
    DRAW_CALLFLAG: _("Because both players ran out of time"),
    DRAW_STALEMATE: _("Because %(mover)s stalemated"),
    DRAW_AGREE: _("Because both players agreed to a draw"),
    DRAW_ADJUDICATION: _("Because of adjudication by an admin"),
    DRAW_LENGTH: _("Because the game exceed the max length"),
    DRAW_BLACKINSUFFICIENTANDWHITETIME:
    _("Because %(white)s ran out of time and %(black)s has insufficient material to mate"),
    DRAW_WHITEINSUFFICIENTANDBLACKTIME:
    _("Because %(black)s ran out of time and %(white)s has insufficient material to mate"),
    DRAW_EQUALMATERIAL:
    _("Because both players have the same amount of pieces"),
    DRAW_KINGSINEIGHTROW: _("Because both king reached the eight row"),
    WON_RESIGN: _("Because %(loser)s resigned"),
    WON_CALLFLAG: _("Because %(loser)s ran out of time"),
    WON_MATE: _("Because %(loser)s was checkmated"),
    WON_DISCONNECTION: _("Because %(loser)s disconnected"),
    WON_ADJUDICATION: _("Because of adjudication by an admin"),
    WON_LESSMATERIAL: _("Because %(winner)s has fewer pieces"),
    WON_NOMATERIAL: _("Because %(winner)s lost all pieces"),
    WON_KINGEXPLODE: _("Because %(loser)s king exploded"),
    WON_KINGINCENTER: _("Because %(winner)s king reached the center"),
    WON_THREECHECK: _("Because %(winner)s was giving check 3 times"),
    WON_KINGINEIGHTROW: _("Because %(winner)s king reached the eight row"),
    WON_WIPEOUT: _("Because %(winner)s wiped out white horde"),
    ADJOURNED_LOST_CONNECTION: _("Because a player lost connection"),
    ADJOURNED_AGREEMENT: _("Because both players agreed to an adjournment"),
    ADJOURNED_SERVER_SHUTDOWN: _("Because the server was shut down"),
    ADJOURNED_COURTESY:
    _("Because a player lost connection and the other player requested adjournment"),
    ADJOURNED_COURTESY_WHITE:
    _("Because %(black)s lost connection to the server and %(white)s requested adjournment"),
    ADJOURNED_COURTESY_BLACK:
    _("Because %(white)s lost connection to the server and %(black)s requested adjournment"),
    ADJOURNED_LOST_CONNECTION_WHITE:
    _("Because %(white)s lost connection to the server"),
    ADJOURNED_LOST_CONNECTION_BLACK:
    _("Because %(black)s lost connection to the server"),
    ABORTED_ADJUDICATION:
    _("Because of adjudication by an admin. No rating changes have occurred."),
    ABORTED_AGREEMENT:
    _("Because both players agreed to abort the game. No rating changes have occurred."),
    ABORTED_COURTESY:
    _("Because of courtesy by a player. No rating changes have occurred."),
    ABORTED_EARLY:
    _("Because a player aborted the game. Either player can abort the game without \
      the other's consent before the second move. No rating changes have occurred."),
    ABORTED_DISCONNECTION:
    _("Because a player disconnected and there are too few moves to warrant adjournment. No rating changes have occurred."),
    ABORTED_SERVER_SHUTDOWN:
    _("Because the server was shut down. No rating changes have occurred."),
    WHITE_ENGINE_DIED: _("Because the %(white)s engine died"),
    BLACK_ENGINE_DIED: _("Because the %(black)s engine died"),
    DISCONNECTED: _("Because the connection to the server was lost"),
    UNKNOWN_REASON: _("The reason is unknown"),
    PRACTICE_GOAL_REACHED: _("Because practice goal reached"),
}



import io
import os
import sys
import subprocess
from urllib.request import url2pathname

from pychess.System.Log import log


class Player():
    def __init__(self):
        self.ready = False

    def play(self, uri):
        pass


sound_player = Player()

if sys.platform == "win32":
    import winsound

    class WinsoundPlayer(Player):
        def __init__(self):
            self.ready = True

        def play(self, uri):
            try:
                winsound.PlaySound(None, 0)
                winsound.PlaySound(
                    url2pathname(uri[5:]), winsound.SND_FILENAME | winsound.SND_ASYNC)
            except RuntimeError:
                log.error("ERROR: RuntimeError while playing %s." %
                          url2pathname(uri[5:]))

    sound_player = WinsoundPlayer()
else:
    class GstPlayer(Player):
        def __init__(self):
            PYTHONBIN = sys.executable.split("/")[-1]
            try:
                if getattr(sys, 'frozen', False):
                    gst_player = os.path.join(os.path.abspath(os.path.dirname(sys.executable)), "gst_player.py")
                else:
                    gst_player = os.path.join(os.path.abspath(os.path.dirname(__file__)), "gst_player.py")
                self.player = subprocess.Popen([PYTHONBIN, gst_player],
                                               stdin=subprocess.PIPE)

                self.stdin = io.TextIOWrapper(self.player.stdin, encoding='utf-8', line_buffering=True)
                self.ready = True
            except Exception:
                self.player = None
                log.error('ERROR: starting gst_player failed')
                raise

        def play(self, uri):
            if self.player is not None:
                try:
                    self.stdin.write(url2pathname(uri[5:]) + "\n")
                except BrokenPipeError:
                    pass

    sound_player = GstPlayer()



from pychess.Utils.const import VARIANTS_BLINDFOLD, BLINDFOLDCHESS, HIDDENPAWNSCHESS, \
    HIDDENPIECESCHESS, ALLWHITECHESS
from pychess.Utils.Board import Board


class BlindfoldBoard(Board):
    variant = BLINDFOLDCHESS
    __desc__ = _("Classic chess rules with hidden figurines\n" +
                 "http://en.wikipedia.org/wiki/Blindfold_chess")
    name = _("Blindfold")
    cecp_name = "normal"
    need_initial_board = False
    standard_rules = True
    variant_group = VARIANTS_BLINDFOLD


class HiddenPawnsBoard(Board):
    variant = HIDDENPAWNSCHESS
    __desc__ = _("Classic chess rules with hidden pawns\n" +
                 "http://en.wikipedia.org/wiki/Blindfold_chess")
    name = _("Hidden pawns")
    cecp_name = "normal"
    need_initial_board = False
    standard_rules = True
    variant_group = VARIANTS_BLINDFOLD


class HiddenPiecesBoard(Board):
    variant = HIDDENPIECESCHESS
    __desc__ = _("Classic chess rules with hidden pieces\n" +
                 "http://en.wikipedia.org/wiki/Blindfold_chess")
    name = _("Hidden pieces")
    cecp_name = "normal"
    need_initial_board = False
    standard_rules = True
    variant_group = VARIANTS_BLINDFOLD


class AllWhiteBoard(Board):
    variant = ALLWHITECHESS
    __desc__ = _("Classic chess rules with all pieces white\n" +
                 "http://en.wikipedia.org/wiki/Blindfold_chess")
    name = _("All white")
    cecp_name = "normal"
    need_initial_board = False
    standard_rules = True
    variant_group = VARIANTS_BLINDFOLD



from .bitboard import bitPosArray, iterBits, clearBit, firstBit
from .attack import isAttacked, pinnedOnKing, getAttacks
from .ldata import fromToRay, moveArray, directions, fileBits, rankBits,\
    ray45, attack45, ray135, attack135, ray90, attack90, ray00, attack00, FILE, rays
from pychess.Utils.const import EMPTY, PAWN,\
    QUEEN, KNIGHT, BISHOP, ROOK, KING, HAWK, ELEPHANT, WHITE, BLACK,\
    SITTUYINCHESS, FISCHERRANDOMCHESS, SUICIDECHESS, GIVEAWAYCHESS, CAMBODIANCHESS,\
    ATOMICCHESS, WILDCASTLECHESS, WILDCASTLESHUFFLECHESS, CRAZYHOUSECHESS, ASEAN_VARIANTS,\
    HORDECHESS, PLACEMENTCHESS, SCHESS, LIGHTBRIGADECHESS, BPAWN, sliders,\
    A8, A6, G6, F6, H1, C3, B2, B3, A3, D6, D8, E3, E1, E8, C7, F2, D1, E6, H3, D3, H2, G7, H6, H7,\
    ASEAN_QUEEN, ASEAN_BBISHOP, ASEAN_WBISHOP, NORMAL_MOVE, QUEEN_CASTLE, KING_CASTLE, ENPASSANT,\
    KNIGHT_PROMOTION, BISHOP_PROMOTION, ROOK_PROMOTION, QUEEN_PROMOTION, KING_PROMOTION, \
    DROP_VARIANTS, DROP, B_OOO, B_OO, W_OOO, W_OO, HAWK_PROMOTION, ELEPHANT_PROMOTION, \
    HAWK_GATE, ELEPHANT_GATE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK

# The format of a move is as follows - from left:
# 4 bits:  Descriping the type of the move
# 6 bits:  cord to move from
# 6 bits:  cord to move to

shiftedFromCords = []
for i in range(64):
    shiftedFromCords.append(i << 6)

shiftedFlags = []
for i in NORMAL_MOVE, QUEEN_CASTLE, KING_CASTLE, ENPASSANT, \
        KNIGHT_PROMOTION, BISHOP_PROMOTION, ROOK_PROMOTION, QUEEN_PROMOTION, \
        KING_PROMOTION, HAWK_PROMOTION, ELEPHANT_PROMOTION, \
        HAWK_GATE, ELEPHANT_GATE, HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK, DROP:
    shiftedFlags.append(i << 12)


def newMove(fromcord, tocord, flag=NORMAL_MOVE):
    return shiftedFlags[flag] + shiftedFromCords[fromcord] + tocord

# Generate all moves


def genCastles(board):
    def generateOne(color, side, king_after, rook_after):
        if side == 0:
            castle = QUEEN_CASTLE
        else:
            castle = KING_CASTLE
        king = board.ini_kings[color]
        rook = board.ini_rooks[color][side]
        blocker = clearBit(clearBit(board.blocker, king), rook)
        stepover = fromToRay[king][king_after] | fromToRay[rook][rook_after]
        if not stepover & blocker:
            for cord in range(
                    min(king, king_after), max(king, king_after) + 1):
                if isAttacked(board, cord, 1 - color):
                    return
            if FILE(king) == 3 and board.variant in (WILDCASTLECHESS,
                                                     WILDCASTLESHUFFLECHESS):
                castle = QUEEN_CASTLE if castle == KING_CASTLE else KING_CASTLE
            if board.variant == FISCHERRANDOMCHESS:
                yield newMove(king, rook, castle)
            else:
                yield newMove(king, king_after, castle)
                if board.variant == SCHESS:
                    holding = board.holding[color]
                    if holding[HAWK] == 0 and holding[ELEPHANT] == 0:
                        return
                    for king in iterBits(board.boards[board.color][KING] & board.virgin[board.color]):
                        if holding[HAWK] > 0:
                            yield newMove(king, king_after, HAWK_GATE)
                            yield newMove(rook, king, HAWK_GATE_AT_ROOK)
                        if holding[ELEPHANT] > 0:
                            yield newMove(king, king_after, ELEPHANT_GATE)
                            yield newMove(rook, king, ELEPHANT_GATE_AT_ROOK)

    king = board.ini_kings[board.color]
    wildcastle = FILE(king) == 3 and board.variant in (WILDCASTLECHESS,
                                                       WILDCASTLESHUFFLECHESS)
    if board.color == WHITE:
        if board.castling & W_OO:
            side = 0 if wildcastle else 1
            for move in generateOne(WHITE, side, board.fin_kings[WHITE][side],
                                    board.fin_rooks[WHITE][side]):
                yield move

        if board.castling & W_OOO:
            side = 1 if wildcastle else 0
            for move in generateOne(WHITE, side, board.fin_kings[WHITE][side],
                                    board.fin_rooks[WHITE][side]):
                yield move
    else:
        if board.castling & B_OO:
            side = 0 if wildcastle else 1
            for move in generateOne(BLACK, side, board.fin_kings[BLACK][side],
                                    board.fin_rooks[BLACK][side]):
                yield move

        if board.castling & B_OOO:
            side = 1 if wildcastle else 0
            for move in generateOne(BLACK, side, board.fin_kings[BLACK][side],
                                    board.fin_rooks[BLACK][side]):
                yield move


def genPieceMoves(board, piece, tcord):
    """"
    Used by parseSAN only to accelerate it a bit
    """
    moves = set()
    friends = board.friends[board.color]
    notfriends = ~friends
    if piece == KNIGHT:
        knights = board.boards[board.color][KNIGHT]
        knightMoves = moveArray[KNIGHT]
        for fcord in iterBits(knights):
            if tcord in iterBits(knightMoves[fcord] & notfriends):
                moves.add(newMove(fcord, tcord))
        return moves

    if piece == BISHOP:
        bishops = board.boards[board.color][BISHOP]
        if board.variant in ASEAN_VARIANTS:
            bishopMoves = moveArray[ASEAN_WBISHOP if board.color == WHITE else
                                    ASEAN_BBISHOP]
            for fcord in iterBits(bishops):
                if tcord in iterBits(bishopMoves[fcord] & notfriends):
                    moves.add(newMove(fcord, tcord))
            return moves
        else:
            blocker = board.blocker
            for fcord in iterBits(bishops):
                try:
                    attackBoard = attack45[fcord][ray45[fcord] & blocker] | \
                        attack135[fcord][ray135[fcord] & blocker]
                except KeyError:
                    attackBoard = 0
                if tcord in iterBits(attackBoard & notfriends):
                    moves.add(newMove(fcord, tcord))
            return moves

    if piece == ROOK:
        blocker = board.blocker
        rooks = board.boards[board.color][ROOK]
        for fcord in iterBits(rooks):
            try:
                attackBoard = attack00[fcord][ray00[fcord] & blocker] | \
                    attack90[fcord][ray90[fcord] & blocker]
            except KeyError:
                attackBoard = 0
            if tcord in iterBits(attackBoard & notfriends):
                moves.add(newMove(fcord, tcord))
        return moves

    if piece == QUEEN:
        queens = board.boards[board.color][QUEEN]
        if board.variant in ASEAN_VARIANTS:
            queenMoves = moveArray[ASEAN_QUEEN]
            for fcord in iterBits(queens):
                if tcord in iterBits(queenMoves[fcord] & notfriends):
                    moves.add(newMove(fcord, tcord))
            # Cambodian extra first move
            if board.variant == CAMBODIANCHESS:
                if board.is_first_move[QUEEN][board.color]:
                    if board.color == WHITE:
                        if not board.arBoard[E3]:
                            moves.add(newMove(E1, E3))
                    else:
                        if not board.arBoard[D6]:
                            moves.add(newMove(D8, D6))
            return moves
        else:
            blocker = board.blocker
            for fcord in iterBits(queens):
                try:
                    attackBoard = attack45[fcord][ray45[fcord] & blocker] | \
                        attack135[fcord][ray135[fcord] & blocker]
                except KeyError:
                    attackBoard = 0
                if tcord in iterBits(attackBoard & notfriends):
                    moves.add(newMove(fcord, tcord))

                try:
                    attackBoard = attack00[fcord][ray00[fcord] & blocker] | \
                        attack90[fcord][ray90[fcord] & blocker]
                except KeyError:
                    attackBoard = 0
                if tcord in iterBits(attackBoard & notfriends):
                    moves.add(newMove(fcord, tcord))
            return moves

    if (board.variant == SUICIDECHESS or board.variant == GIVEAWAYCHESS) and piece == KING:
        kings = board.boards[board.color][KING]
        if kings:
            kingMoves = moveArray[KING]
            for fcord in iterBits(kings):
                for tc in iterBits(kingMoves[fcord] & notfriends):
                    if tc == tcord:
                        moves.add(newMove(fcord, tcord))
            return moves

    return moves


def gen_sittuyin_promotions(board):
    from pychess.Variants import variants
    blocker = board.blocker
    notblocker = ~blocker

    pawns = board.boards[board.color][PAWN]

    queenMoves = moveArray[ASEAN_QUEEN]

    def willDirectAttack(board, move, cord):
        board_clone = board.clone()
        board_clone.applyMove(move)
        return board.friends[1 - board.color] & moveArray[ASEAN_QUEEN][cord]

    promotion_zone = variants[SITTUYINCHESS].PROMOTION_ZONE[board.color]
    for cord in iterBits(pawns):
        if board.pieceCount[board.color][PAWN] == 1 or cord in promotion_zone:
            # in place promotions
            move = newMove(cord, cord, QUEEN_PROMOTION)
            if not board.willGiveCheck(move) and not willDirectAttack(board, move, cord):
                yield move

            # queen move promotion
            for c in iterBits(queenMoves[cord] & notblocker):
                move = newMove(cord, c, QUEEN_PROMOTION)
                if not board.willGiveCheck(move) and not willDirectAttack(board, move, c):
                    yield move


def genAllMoves(board, drops=True):
    from pychess.Variants import variants
    if drops and board.variant in DROP_VARIANTS and board.variant != SCHESS:
        for move in genDrops(board):
            yield move

    # In sittuyin you have to place your pieces before any real move
    if board.variant == SITTUYINCHESS or board.variant == PLACEMENTCHESS:
        if board.plyCount < 16:
            return

    blocker = board.blocker
    notblocker = ~blocker
    enpassant = board.enpassant

    friends = board.friends[board.color]
    notfriends = ~friends
    enemies = board.friends[1 - board.color]

    pawns = board.boards[board.color][PAWN]
    knights = board.boards[board.color][KNIGHT]
    bishops = board.boards[board.color][BISHOP]
    rooks = board.boards[board.color][ROOK]
    queens = board.boards[board.color][QUEEN]
    kings = board.boards[board.color][KING]
    hawks = board.boards[board.color][HAWK]
    elephants = board.boards[board.color][ELEPHANT]

    PROMOTIONS = variants[board.variant].PROMOTIONS
    # In sittuyin only one queen allowed to exist any time per side
    if board.variant == SITTUYINCHESS and queens:
        PROMOTIONS = (NORMAL_MOVE, )
    elif board.variant == LIGHTBRIGADECHESS:
        PROMOTIONS = (QUEEN_PROMOTION, ) if board.color == WHITE else (KNIGHT_PROMOTION, )

    # Knights, Hawks, Elephants
    knightMoves = moveArray[KNIGHT]
    for cord in iterBits(knights | hawks | elephants):
        for c in iterBits(knightMoves[cord] & notfriends):
            yield newMove(cord, c)

    # knight moves with gating
    holding = board.holding[board.color]
    if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
        for cord in iterBits(knights & board.virgin[board.color]):
            for c in iterBits(knightMoves[cord] & notfriends):
                if holding[HAWK] > 0:
                    yield newMove(cord, c, HAWK_GATE)
                if holding[ELEPHANT] > 0:
                    yield newMove(cord, c, ELEPHANT_GATE)

    # King
    if kings:
        kingMoves = moveArray[KING]
        # cord = firstBit(kings)
        for cord in iterBits(kings):
            for c in iterBits(kingMoves[cord] & notfriends):
                if board.variant == ATOMICCHESS:
                    if not board.arBoard[c]:
                        yield newMove(cord, c)
                else:
                    yield newMove(cord, c)
                    if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
                        for cord in iterBits(kings & board.virgin[board.color]):
                            if holding[HAWK] > 0:
                                yield newMove(cord, c, HAWK_GATE)
                            if holding[ELEPHANT] > 0:
                                yield newMove(cord, c, ELEPHANT_GATE)

    if board.variant in ASEAN_VARIANTS:
        # Rooks
        for cord in iterBits(rooks):
            try:
                attackBoard = attack00[cord][ray00[cord] & blocker] | \
                    attack90[cord][ray90[cord] & blocker]
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & notfriends):
                yield newMove(cord, c)

        # Queens
        queenMoves = moveArray[ASEAN_QUEEN]
        for cord in iterBits(queens):
            for c in iterBits(queenMoves[cord] & notfriends):
                yield newMove(cord, c)

        # Bishops
        bishopMoves = moveArray[ASEAN_WBISHOP if board.color == WHITE else
                                ASEAN_BBISHOP]
        for cord in iterBits(bishops):
            for c in iterBits(bishopMoves[cord] & notfriends):
                yield newMove(cord, c)

    else:
        # Rooks and Queens and Elephants
        for cord in iterBits(rooks | queens | elephants):
            try:
                attackBoard = attack00[cord][ray00[cord] & blocker] | \
                    attack90[cord][ray90[cord] & blocker]
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & notfriends):
                yield newMove(cord, c)

        if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
            for cord in iterBits((rooks | queens) & board.virgin[board.color]):
                try:
                    attackBoard = attack00[cord][ray00[cord] & blocker] | \
                        attack90[cord][ray90[cord] & blocker]
                except KeyError:
                    attackBoard = 0
                for c in iterBits(attackBoard & notfriends):
                    if holding[HAWK] > 0:
                        yield newMove(cord, c, HAWK_GATE)
                    if holding[ELEPHANT] > 0:
                        yield newMove(cord, c, ELEPHANT_GATE)

    # Bishops and Queens, Hawks
        for cord in iterBits(bishops | queens | hawks):
            try:
                attackBoard = attack45[cord][ray45[cord] & blocker] | \
                    attack135[cord][ray135[cord] & blocker]
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & notfriends):
                yield newMove(cord, c)

        if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
            for cord in iterBits((bishops | queens) & board.virgin[board.color]):
                try:
                    attackBoard = attack45[cord][ray45[cord] & blocker] | \
                        attack135[cord][ray135[cord] & blocker]
                except KeyError:
                    attackBoard = 0
                for c in iterBits(attackBoard & notfriends):
                    if holding[HAWK] > 0:
                        yield newMove(cord, c, HAWK_GATE)
                    if holding[ELEPHANT] > 0:
                        yield newMove(cord, c, ELEPHANT_GATE)

    # White pawns
    pawnEnemies = enemies | (enpassant is not None and bitPosArray[enpassant] or 0)
    if board.color == WHITE:

        # One step

        if board.variant == SITTUYINCHESS:
            promotion_zone = []
        else:
            promotion_zone = variants[board.variant].PROMOTION_ZONE[WHITE]
        movedpawns = (pawns >>
                      8) & notblocker  # Move all pawns one step forward
        for cord in iterBits(movedpawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 8, cord, p)
            else:
                yield newMove(cord - 8, cord)

        # Two steps

        seccondrow = pawns & rankBits[1]  # Get seccond row pawns
        movedpawns = (seccondrow >>
                      8) & notblocker  # Move two steps forward, while
        movedpawns = (movedpawns >>
                      8) & notblocker  # ensuring middle cord is clear
        for cord in iterBits(movedpawns):
            yield newMove(cord - 16, cord)

        # In horde white pawns on first rank may move two squares also
        if board.variant == HORDECHESS:
            firstrow = pawns & rankBits[0]  # Get first row pawns
            movedpawns = (firstrow >>
                          8) & notblocker  # Move two steps forward, while
            movedpawns = (movedpawns >>
                          8) & notblocker  # ensuring middle cord is clear
            for cord in iterBits(movedpawns):
                yield newMove(cord - 16, cord)

        # Capture left

        capLeftPawns = pawns & ~fileBits[0]
        capLeftPawns = (capLeftPawns >> 7) & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 7, cord, ENPASSANT)
            else:
                yield newMove(cord - 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[7]
        capRightPawns = (capRightPawns >> 9) & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 9, cord, ENPASSANT)
            else:
                yield newMove(cord - 9, cord)

    # Black pawns
    else:

        # One step

        if board.variant == SITTUYINCHESS:
            promotion_zone = []
        else:
            promotion_zone = variants[board.variant].PROMOTION_ZONE[BLACK]
        movedpawns = (pawns << 8) & notblocker
        movedpawns &= 0xffffffffffffffff  # contrain to 64 bits
        for cord in iterBits(movedpawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 8, cord, p)
            else:
                yield newMove(cord + 8, cord)

        # Two steps

        seccondrow = pawns & rankBits[6]  # Get seventh row pawns
        # Move two steps forward, while ensuring middle cord is clear
        movedpawns = seccondrow << 8 & notblocker
        movedpawns = movedpawns << 8 & notblocker
        for cord in iterBits(movedpawns):
            yield newMove(cord + 16, cord)

        # Capture left

        capLeftPawns = pawns & ~fileBits[7]
        capLeftPawns = capLeftPawns << 7 & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 7, cord, ENPASSANT)
            else:
                yield newMove(cord + 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[0]
        capRightPawns = capRightPawns << 9 & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 9, cord, ENPASSANT)
            else:
                yield newMove(cord + 9, cord)

    # Sittuyin promotions
    if board.variant == SITTUYINCHESS and pawns and not queens:
        for move in gen_sittuyin_promotions(board):
            yield move

    # Cambodian extra first moves for king and queen
    if board.variant == CAMBODIANCHESS:
        if board.arBoard[board.ini_kings[board.color]] == KING and \
                board.is_first_move[KING][board.color]:
            if board.color == WHITE:
                if not board.arBoard[B2]:
                    yield newMove(D1, B2)
                if not board.arBoard[F2]:
                    yield newMove(D1, F2)
            else:
                if not board.arBoard[C7]:
                    yield newMove(E8, C7)
                if not board.arBoard[G7]:
                    yield newMove(E8, G7)
        if board.arBoard[board.ini_queens[board.color]] == QUEEN and \
                board.is_first_move[QUEEN][board.color]:
            if board.color == WHITE:
                if not board.arBoard[E3]:
                    yield newMove(E1, E3)
            else:
                if not board.arBoard[D6]:
                    yield newMove(D8, D6)

    # Castling
    if kings:
        for move in genCastles(board):
            yield move

################################################################################
#   Generate capturing moves                                                   #
################################################################################


def genCaptures(board):
    from pychess.Variants import variants

    blocker = board.blocker
    enpassant = board.enpassant

    enemies = board.friends[1 - board.color]

    pawns = board.boards[board.color][PAWN]
    knights = board.boards[board.color][KNIGHT]
    bishops = board.boards[board.color][BISHOP]
    rooks = board.boards[board.color][ROOK]
    queens = board.boards[board.color][QUEEN]
    kings = board.boards[board.color][KING]
    hawks = board.boards[board.color][HAWK]
    elephants = board.boards[board.color][ELEPHANT]

    PROMOTIONS = variants[board.variant].PROMOTIONS
    # In sittuyin promotion can't give capture
    if board.variant == SITTUYINCHESS:
        PROMOTIONS = (NORMAL_MOVE, )
    elif board.variant == LIGHTBRIGADECHESS:
        PROMOTIONS = (QUEEN_PROMOTION, ) if board.color == WHITE else (KNIGHT_PROMOTION, )

    # Knights
    knightMoves = moveArray[KNIGHT]
    for cord in iterBits(knights | hawks | elephants):
        for c in iterBits(knightMoves[cord] & enemies):
            yield newMove(cord, c)

    # King
    if kings:
        kingMoves = moveArray[KING]
        # cord = firstBit(kings)
        for cord in iterBits(kings):
            for c in iterBits(kingMoves[cord] & enemies):
                if board.variant != ATOMICCHESS:
                    yield newMove(cord, c)

    # Rooks and Queens
    if board.variant in ASEAN_VARIANTS:
        for cord in iterBits(rooks):
            try:
                attackBoard = attack00[cord][ray00[cord] & blocker] | \
                    attack90[cord][ray90[cord] & blocker]
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & enemies):
                yield newMove(cord, c)
    else:
        for cord in iterBits(rooks | queens | elephants):
            try:
                attackBoard = attack00[cord][ray00[cord] & blocker] | \
                    attack90[cord][ray90[cord] & blocker]
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & enemies):
                yield newMove(cord, c)

    # Bishops and Queens
    if board.variant in ASEAN_VARIANTS:
        bishopMoves = moveArray[ASEAN_WBISHOP if board.color == WHITE else
                                ASEAN_BBISHOP]
        for cord in iterBits(bishops):
            for c in iterBits(bishopMoves[cord] & enemies):
                yield newMove(cord, c)
        queenMoves = moveArray[ASEAN_QUEEN]
        for cord in iterBits(queens):
            for c in iterBits(queenMoves[cord] & enemies):
                yield newMove(cord, c)
    else:
        for cord in iterBits(bishops | queens | hawks):
            try:
                attackBoard = attack45[cord][ray45[cord] & blocker] | \
                    attack135[cord][ray135[cord] & blocker]
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & enemies):
                yield newMove(cord, c)

    # White pawns
    pawnEnemies = enemies | (enpassant is not None and bitPosArray[enpassant] or 0)

    if board.color == WHITE:
        promotion_zone = variants[board.variant].PROMOTION_ZONE[WHITE]

        # Promotes

        # Capture left

        capLeftPawns = pawns & ~fileBits[0]
        capLeftPawns = (capLeftPawns >> 7) & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 7, cord, ENPASSANT)
            else:
                yield newMove(cord - 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[7]
        capRightPawns = (capRightPawns >> 9) & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 9, cord, ENPASSANT)
            else:
                yield newMove(cord - 9, cord)

    # Black pawns
    else:
        promotion_zone = variants[board.variant].PROMOTION_ZONE[BLACK]

        # One step

        # Capture left

        capLeftPawns = pawns & ~fileBits[7]
        capLeftPawns = capLeftPawns << 7 & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 7, cord, ENPASSANT)
            else:
                yield newMove(cord + 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[0]
        capRightPawns = capRightPawns << 9 & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 9, cord, ENPASSANT)
            else:
                yield newMove(cord + 9, cord)

################################################################################
#   Generate escapes from check                                                #
################################################################################


def genCheckEvasions(board):
    from pychess.Variants import variants
    color = board.color
    opcolor = 1 - color

    kcord = board.kings[color]
    kings = board.boards[color][KING]
    pawns = board.boards[color][PAWN]
    queens = board.boards[board.color][QUEEN]
    checkers = getAttacks(board, kcord, opcolor)

    arBoard = board.arBoard
    if bin(checkers).count("1") == 1:

        PROMOTIONS = variants[board.variant].PROMOTIONS
        # In sittuyin promotion move not allowed to capture opponent pieces
        if board.variant == SITTUYINCHESS and board.boards[board.color][QUEEN]:
            PROMOTIONS = (NORMAL_MOVE, )
        elif board.variant == LIGHTBRIGADECHESS:
            PROMOTIONS = (QUEEN_PROMOTION, ) if board.color == WHITE else (KNIGHT_PROMOTION, )
        promotion_zone = variants[board.variant].PROMOTION_ZONE[color]

        # Captures of checking pieces (except by king, which we will test later)
        chkcord = firstBit(checkers)
        b = getAttacks(board, chkcord, color) & ~kings
        for cord in iterBits(b):
            if not pinnedOnKing(board, cord, color):
                if arBoard[cord] == PAWN and chkcord in promotion_zone and board.variant != SITTUYINCHESS:
                    for p in PROMOTIONS:
                        yield newMove(cord, chkcord, p)
                else:
                    yield newMove(cord, chkcord)

        # Maybe enpassant can help
        if board.enpassant:
            ep = board.enpassant
            if ep + (color == WHITE and -8 or 8) == chkcord:
                bits = moveArray[color == WHITE and BPAWN or PAWN][ep] & pawns
                for cord in iterBits(bits):
                    if not pinnedOnKing(board, cord, color):
                        yield newMove(cord, ep, ENPASSANT)

        # Lets block/capture the checking piece
        if sliders[arBoard[chkcord]]:
            bits = clearBit(fromToRay[kcord][chkcord], chkcord)

            for cord in iterBits(bits):
                b = getAttacks(board, cord, color)
                b &= ~(kings | pawns)

                # Add in pawn advances
                if color == WHITE and cord > H2:
                    if bitPosArray[cord - 8] & pawns:
                        b |= bitPosArray[cord - 8]
                    if cord >> 3 == 3 and arBoard[cord - 8] == EMPTY and \
                            bitPosArray[cord - 16] & pawns:
                        b |= bitPosArray[cord - 16]

                elif color == BLACK and cord < H7:
                    if bitPosArray[cord + 8] & pawns:
                        b |= bitPosArray[cord + 8]
                    if cord >> 3 == 4 and arBoard[cord + 8] == EMPTY and \
                            bitPosArray[cord + 16] & pawns:
                        b |= bitPosArray[cord + 16]

                for fcord in iterBits(b):
                    # If the piece is blocking another attack, we cannot move it
                    if pinnedOnKing(board, fcord, color):
                        continue
                    if arBoard[fcord] == PAWN and cord in promotion_zone:
                        for p in PROMOTIONS:
                            yield newMove(fcord, cord, p)
                    else:
                        yield newMove(fcord, cord)

                if board.variant == CRAZYHOUSECHESS:
                    holding = board.holding[color]
                    for piece in holding:
                        if holding[piece] > 0:
                            if piece == PAWN:
                                if cord >= 56 or cord <= 7:
                                    continue
                            yield newMove(piece, cord, DROP)

                if board.variant == SITTUYINCHESS and pawns and not queens:
                    from .lmove import TCORD
                    for move in gen_sittuyin_promotions(board):
                        if TCORD(move) == cord:
                            yield move

    # If more than one checkers, move king to get out of check
    if checkers:
        escapes = moveArray[KING][kcord] & ~board.friends[color]
    else:
        escapes = 0

    for chkcord in iterBits(checkers):
        dir = directions[chkcord][kcord]
        if sliders[arBoard[chkcord]]:
            escapes &= ~rays[chkcord][dir]

    for cord in iterBits(escapes):
        if not isAttacked(board, cord, opcolor):
            yield newMove(kcord, cord)


def genDrops(board):
    color = board.color
    arBoard = board.arBoard
    holding = board.holding[color]
    for piece in holding:
        if holding[piece] > 0:
            for cord, elem in enumerate(arBoard):
                if elem == EMPTY:
                    # forbidden drop moves
                    if board.variant == SITTUYINCHESS:
                        if color == WHITE:
                            if cord in (A3, B3, C3, D3) or cord > H3:
                                continue
                            if piece == ROOK and cord > H1:
                                continue
                        else:
                            if cord in (E6, F6, G6, H6) or cord < A6:
                                continue
                            if piece == ROOK and cord < A8:
                                continue

                    elif board.variant == PLACEMENTCHESS:
                        # drop pieces enabled on base line only
                        if color == WHITE:
                            if cord > H1:
                                continue
                        else:
                            if cord < A8:
                                continue

                        # bishops must be on opposite colour squares
                        base_line = arBoard[0:8] if color == WHITE else arBoard[56:64]
                        occupied_colors = [0, 0]
                        occupied_colors[cord % 2] += 1
                        for i, baseline_piece in enumerate(base_line):
                            if baseline_piece != EMPTY:
                                occupied_colors[i % 2] += 1

                        if holding[BISHOP] == 2 and piece != BISHOP:
                            # occupying all same colored fields before any bishop dropped is no-no
                            if occupied_colors[WHITE] == 4 or occupied_colors[BLACK] == 4:
                                continue
                        elif holding[BISHOP] == 1:
                            for i, baseline_piece in enumerate(base_line):
                                if baseline_piece == BISHOP:
                                    first_bishop_cord = i
                                    break
                            # occupying all possible place of opp colored bishop is no-no
                            if piece != BISHOP and occupied_colors[1 - first_bishop_cord % 2] == 4:
                                continue
                            # same colored bishop is no-no
                            elif piece == BISHOP and first_bishop_cord % 2 == cord % 2:
                                continue

                    if piece == PAWN:
                        if cord >= 56 or cord <= 7:
                            continue

                    yield newMove(piece, cord, DROP)



from gi.repository import GObject

sanmove = "([a-hx@OoPKQRBN0-8+#=-]{2,7})"


class ErrorManager(GObject.GObject):

    __gsignals__ = {
        'onCommandNotFound': (GObject.SignalFlags.RUN_FIRST, None, (str, )),
        'onAmbiguousMove': (GObject.SignalFlags.RUN_FIRST, None, (str, )),
        'onIllegalMove': (GObject.SignalFlags.RUN_FIRST, None, (str, )),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        connection.expect_line(self.onError, r"(.*?): Command not found\.")
        connection.expect_line(self.onAmbiguousMove,
                               r"Ambiguous move \((%s)\)\." % sanmove)
        connection.expect_line(self.onIllegalMove, r"Illegal move \((%s)\)\." %
                               sanmove)

    def onError(self, match):
        command = match.groups()[0]
        self.emit("onCommandNotFound", command)

    def onAmbiguousMove(self, match):
        move = match.groups()[0]
        self.emit("onAmbiguousMove", move)

    def onIllegalMove(self, match):
        move = match.groups()[0]
        self.emit("onIllegalMove", move)
